# Copyright (c) 2025 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from typing import Iterable, Optional, Tuple

import torch
import torch.nn as nn

from tico.quantization.config.ptq import PTQConfig
from tico.quantization.wrapq.wrappers.ptq_wrapper import PTQWrapper
from tico.quantization.wrapq.wrappers.quant_module_base import QuantModuleBase
from tico.quantization.wrapq.wrappers.registry import try_register


@try_register(
    "transformers.models.qwen3_vl.modeling_qwen3_vl.Qwen3VLTextAttention",
)
class QuantQwen3VLTextAttention(QuantModuleBase):
    def __init__(
        self,
        fp_attn: nn.Module,
        *,
        qcfg: Optional[PTQConfig] = None,
        fp_name: Optional[str] = None,
    ):
        super().__init__(qcfg, fp_name=fp_name)

        cfg = fp_attn.config
        self.config = cfg
        self.layer_idx = getattr(fp_attn, "layer_idx", None)

        # Head shapes
        assert hasattr(cfg, "hidden_size") and hasattr(cfg, "num_attention_heads")
        assert hasattr(cfg, "num_key_value_heads")
        self.head_dim = getattr(
            cfg, "head_dim", cfg.hidden_size // cfg.num_attention_heads
        )
        self.kv_rep = cfg.num_attention_heads // cfg.num_key_value_heads

        # --- Wrap projection + norms via PTQWrapper --------------------------
        q_cfg = qcfg.child("q_proj") if qcfg else None
        k_cfg = qcfg.child("k_proj") if qcfg else None
        v_cfg = qcfg.child("v_proj") if qcfg else None
        o_cfg = qcfg.child("o_proj") if qcfg else None
        qn_cfg = qcfg.child("q_norm") if qcfg else None
        kn_cfg = qcfg.child("k_norm") if qcfg else None

        assert hasattr(fp_attn, "q_proj") and isinstance(fp_attn.q_proj, nn.Module)
        assert hasattr(fp_attn, "k_proj") and isinstance(fp_attn.k_proj, nn.Module)
        assert hasattr(fp_attn, "v_proj") and isinstance(fp_attn.v_proj, nn.Module)
        assert hasattr(fp_attn, "o_proj") and isinstance(fp_attn.o_proj, nn.Module)
        assert hasattr(fp_attn, "q_norm") and isinstance(fp_attn.q_norm, nn.Module)
        assert hasattr(fp_attn, "k_norm") and isinstance(fp_attn.k_norm, nn.Module)

        self.q_proj = PTQWrapper(
            fp_attn.q_proj, qcfg=q_cfg, fp_name=f"{fp_name}.q_proj"
        )
        self.k_proj = PTQWrapper(
            fp_attn.k_proj, qcfg=k_cfg, fp_name=f"{fp_name}.k_proj"
        )
        self.v_proj = PTQWrapper(
            fp_attn.v_proj, qcfg=v_cfg, fp_name=f"{fp_name}.v_proj"
        )
        self.o_proj = PTQWrapper(
            fp_attn.o_proj, qcfg=o_cfg, fp_name=f"{fp_name}.o_proj"
        )
        self.q_norm = PTQWrapper(
            fp_attn.q_norm, qcfg=qn_cfg, fp_name=f"{fp_name}.q_norm"
        )
        self.k_norm = PTQWrapper(
            copy.deepcopy(fp_attn.k_norm), qcfg=kn_cfg, fp_name=f"{fp_name}.k_norm"
        )

        # Constant scale (1/√d)
        scale_t = torch.tensor(
            float(getattr(fp_attn, "scaling", self.head_dim**-0.5))
        )
        # Merge scale_t to k_norm, (otherwise merge it to q_norm)
        with torch.no_grad():
            self.k_norm.wrapped.module.weight.mul_(scale_t)

        mk = self._make_obs
        self.obs_hidden = mk("hidden")

        # RoPE tables
        self.obs_cos = mk("cos")
        self.obs_sin = mk("sin")

        # rotate_half sub-steps (q)
        self.obs_q_x1 = mk("q_x1")
        self.obs_q_x2 = mk("q_x2")
        self.obs_q_neg = mk("q_neg")
        self.obs_q_cat = mk("q_cat")

        # rotate_half sub-steps (k)
        self.obs_k_x1 = mk("k_x1")
        self.obs_k_x2 = mk("k_x2")
        self.obs_k_neg = mk("k_neg")
        self.obs_k_cat = mk("k_cat")

        # RoPE combine
        self.obs_q_cos = mk("q_cos")
        self.obs_q_sin = mk("q_sin")
        self.obs_q_rot = mk("q_rot")
        self.obs_k_cos = mk("k_cos")
        self.obs_k_sin = mk("k_sin")
        self.obs_k_rot = mk("k_rot")

        # Masking & attention math
        self.obs_causal_mask = mk("causal_mask")
        self.obs_logits = mk("logits")
        self.obs_mask_add = mk("mask_add")
        self.obs_softmax = mk("softmax")
        self.obs_attn_out = mk("attn_out")
        self.obs_attn_weights = mk("attn_weights")
        self.obs_attn_out_h = mk("attn_out_h")

        # Static causal mask template
        assert hasattr(cfg, "max_position_embeddings")
        max_seq = cfg.max_position_embeddings
        mask = torch.full((1, 1, max_seq, max_seq), float("-120"))  # type: ignore[arg-type]
        mask.triu_(1)
        self.register_buffer("causal_mask_template", mask, persistent=False)

    def _rot(self, t, o_x1, o_x2, o_neg, o_cat):
        """rotate_half(t): [-x2, x1] along the last dimension."""
        x1, x2 = torch.chunk(t, 2, dim=-1)
        x1 = self._fq(x1, o_x1)
        x2 = self._fq(x2, o_x2)
        x2n = self._fq(-x2, o_neg)
        return self._fq(torch.cat((x2n, x1), dim=-1), o_cat)

    def _apply_rope(self, q, k, cos, sin, unsqueeze_dim: int = 1):
        cos_u = cos.unsqueeze(unsqueeze_dim)
        sin_u = sin.unsqueeze(unsqueeze_dim)

        q_half = self._rot(
            q, self.obs_q_x1, self.obs_q_x2, self.obs_q_neg, self.obs_q_cat
        )
        q_cos = self._fq(q * cos_u, self.obs_q_cos)
        q_sin = self._fq(q_half * sin_u, self.obs_q_sin)
        q_rot = self._fq(q_cos + q_sin, self.obs_q_rot)

        k_half = self._rot(
            k, self.obs_k_x1, self.obs_k_x2, self.obs_k_neg, self.obs_k_cat
        )
        k_cos = self._fq(k * cos_u, self.obs_k_cos)
        k_sin = self._fq(k_half * sin_u, self.obs_k_sin)
        k_rot = self._fq(k_cos + k_sin, self.obs_k_rot)

        return q_rot, k_rot

    def forward(
        self,
        hidden_states: torch.Tensor,
        position_embeddings: Tuple[torch.Tensor, torch.Tensor],
        attention_mask: Optional[torch.Tensor] = None,
        past_key_values=None,  # HF Cache-like object or None
        cache_position: Optional[torch.LongTensor] = None,
        **kwargs,
    ):
        hidden = self._fq(hidden_states, self.obs_hidden)
        B, S, _ = hidden.shape
        H = self.head_dim

        # Projections
        q = self.q_proj(hidden).view(B, S, -1, H).transpose(1, 2)  # (B, n_h,  S, H)
        k = self.k_proj(hidden).view(B, S, -1, H).transpose(1, 2)  # (B, n_kv, S, H)
        v = self.v_proj(hidden).view(B, S, -1, H).transpose(1, 2)  # (B, n_kv, S, H)

        # Head-dim RMSNorms
        q = self.q_norm(q)
        k = self.k_norm(k)

        # RoPE tables
        cos, sin = position_embeddings
        cos = self._fq(cos, self.obs_cos)
        sin = self._fq(sin, self.obs_sin)
        q_rot, k_rot = self._apply_rope(q, k, cos, sin, unsqueeze_dim=1)

        # TODO Reivist cache logic
        # Cache update (HF Cache path)
        if past_key_values is not None:
            cache_kwargs = {"sin": sin, "cos": cos, "cache_position": cache_position}
            k_rot, v = past_key_values.update(k_rot, v, self.layer_idx, cache_kwargs)

        # Build causal mask if needed
        if attention_mask is None or attention_mask.dtype == torch.bool:
            q_len = q_rot.size(-2)
            k_len = k_rot.size(-2)
            attention_mask = self.causal_mask_template[..., :q_len, :k_len].to(
                hidden.device
            )
        attention_mask = self._fq(attention_mask, self.obs_causal_mask)

        attn_weights_parts = []
        attn_out_parts = []

        n_kv = k_rot.size(1)  # num_key_value_heads
        kv_rep = self.kv_rep

        # TODO Consider attaching a separate observer to each computation.
        for i in range(n_kv):
            # (B, 1, K, H)
            k_i = k_rot[:, i : i + 1, :, :]
            v_i = v[:, i : i + 1, :, :]

            # (B, G, S, H) where G=kv_rep
            h0 = i * kv_rep
            h1 = (i + 1) * kv_rep
            q_i = q_rot[:, h0:h1, :, :]

            # logits: (B, G, S, K)
            logits_i = self._fq(q_i @ k_i.transpose(-2, -1), self.obs_logits)

            # mask add: broadcast on head axis (1 -> G)
            logits_i = self._fq(logits_i + attention_mask, self.obs_mask_add)

            # softmax
            attn_i = torch.softmax(logits_i, dim=-1, dtype=torch.float32).to(
                q_rot.dtype
            )
            attn_i = self._fq(attn_i, self.obs_softmax)

            # out: (B, G, S, H)
            out_i = self._fq(attn_i @ v_i, self.obs_attn_out)

            attn_weights_parts.append(attn_i)
            attn_out_parts.append(out_i)

        # concat heads back: (B, n_h, S, K) / (B, n_h, S, H)
        attn_weights = self._fq(
            torch.cat(attn_weights_parts, dim=1), self.obs_attn_weights
        )
        attn_out_h = self._fq(torch.cat(attn_out_parts, dim=1), self.obs_attn_out_h)

        # Attention output
        attn_out = attn_out_h.transpose(1, 2).reshape(B, S, -1)  # (B, S, n_h * H)

        # Final projection
        out = self.o_proj(attn_out)

        return out, attn_weights

    def _all_observers(self) -> Iterable:
        # local first
        yield from (
            self.obs_hidden,
            self.obs_cos,
            self.obs_sin,
            self.obs_q_x1,
            self.obs_q_x2,
            self.obs_q_neg,
            self.obs_q_cat,
            self.obs_k_x1,
            self.obs_k_x2,
            self.obs_k_neg,
            self.obs_k_cat,
            self.obs_q_cos,
            self.obs_q_sin,
            self.obs_q_rot,
            self.obs_k_cos,
            self.obs_k_sin,
            self.obs_k_rot,
            self.obs_causal_mask,
            self.obs_logits,
            self.obs_mask_add,
            self.obs_softmax,
            self.obs_attn_out,
            self.obs_attn_weights,
            self.obs_attn_out_h,
        )

        for m in (
            self.q_proj,
            self.k_proj,
            self.v_proj,
            self.o_proj,
            self.q_norm,
            self.k_norm,
        ):
            yield from m._all_observers()



# Copyright (c) 2025 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tempfile
from pathlib import Path
from typing import List

import numpy as np
import torch
from circle_schema import circle

from tico.quantization.evaluation.executor.backend_executor import BackendExecutor
from tico.quantization.evaluation.utils import (
    dequantize,
    get_graph_input_output,
    quantize,
)
from tico.serialize.circle_mapping import np_dtype_from_circle_dtype
from tico.utils.model import CircleModel
from tico.utils.utils import run_bash_cmd


class Triv24Executor(BackendExecutor):
    """
    Implementation for a TRIV24 backend.
    """

    def __init__(self):
        self.compiler_path = Path("/usr/share/one/backends/triv24/bin/triv24-compile")
        self.interpreter_path = Path(
            "/usr/share/one/backends/triv24/bin/triv24-ssinfer"
        )
        self.circle_model = None
        self.tvn_path = None

        # Check if triv24 toolchain is installed.
        if not self.compiler_path.is_file() or not self.interpreter_path.is_file():
            raise RuntimeError(
                "Not found triv24 toolchain. Please install the toolchain package first."
            )

        self.temp_dir = tempfile.TemporaryDirectory()

    def compile(self, circle_model: CircleModel) -> None:
        assert isinstance(circle_model, CircleModel)
        self.circle_model = circle_model
        circle_path = Path(self.temp_dir.name) / "quantized.circle"
        circle_model.save(str(circle_path))
        self.tvn_path = Path(self.temp_dir.name) / "compiled.tvn"
        args = []
        args += ["-o", str(self.tvn_path)]
        args += [str(circle_path)]
        cmd = [str(self.compiler_path)] + args
        run_bash_cmd(cmd)

    def run_inference(self, input_data: List[torch.Tensor]) -> List[np.ndarray]:
        if not self.tvn_path:
            raise RuntimeError("You must compile the model before running inference.")

        assert isinstance(self.circle_model, CircleModel)
        circle_inputs, circle_outputs = get_graph_input_output(self.circle_model)
        # Get input/output of scale/zp from quantized circle.
        # Note that qparams may be None because some of them are not quantized like indices, argmax.
        input_qparam: List[circle.QuantizationParameters.QuantizationParameters] = []
        for inp in circle_inputs:
            input_qparam.append(inp.Quantization())
        output_qparam: List[circle.QuantizationParameters.QuantizationParameters] = []
        for out in circle_outputs:
            output_qparam.append(out.Quantization())

        # Create input files for inference.
        for in_idx, data in enumerate(input_data):
            in_data_path = Path(self.temp_dir.name) / f"input.{in_idx}.tv2b"
            assert isinstance(data, torch.Tensor)
            if input_qparam[in_idx] is None:
                np_data = data.numpy()
            else:
                np_data = quantize(
                    data.numpy(),
                    input_qparam[in_idx].ScaleAsNumpy()[0],
                    input_qparam[in_idx].ZeroPointAsNumpy()[0],
                    np_dtype_from_circle_dtype(circle_inputs[in_idx].Type()),
                )
            np_data.tofile(in_data_path)
        args = []
        args += ["--loadable", str(self.tvn_path)]
        args += ["--input-spec", f"tv2b:{str(self.temp_dir.name)}/input"]
        args += ["--dump-output-as-tv2b", f"{str(self.temp_dir.name)}/output"]
        cmd = [str(self.interpreter_path)] + args

        # Run inference
        run_bash_cmd(cmd)

        # Load outputs from file
        circle_dequantized_output = []
        circle_output_shapes_np = [t.ShapeAsNumpy() for t in circle_outputs]
        for out_idx, out_shape in enumerate(circle_output_shapes_np):
            out_data_path = Path(self.temp_dir.name) / f"output.{out_idx}.tv2b"
            circle_output = np.fromfile(out_data_path, np.uint8).reshape(out_shape)
            if output_qparam[out_idx] is None:
                circle_dequantized_output.append(circle_output)
            else:
                circle_dequantized_output.append(
                    dequantize(
                        circle_output,
                        output_qparam[out_idx].ScaleAsNumpy()[0],
                        output_qparam[out_idx].ZeroPointAsNumpy()[0],
                        np_dtype_from_circle_dtype(circle_outputs[out_idx].Type()),
                    )
                )

        return circle_dequantized_output

    def __del__(self):
        self.temp_dir.cleanup()



# Copyright (c) 2025 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob
from os.path import basename, dirname, isfile, join

from tico.utils.register_custom_op import RegisterOps


# Register custom ops to torch namespace
RegisterOps()

# Load all modules in the current directory
modules = glob.glob(join(dirname(__file__), "*.py"))
__all__ = [
    basename(f)[:-3] for f in modules if isfile(f) and not f.endswith("__init__.py")
]



# Copyright (c) 2025 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch

from tico.quantization.wrapq.observers.affine_base import AffineObserverBase
from tico.quantization.wrapq.utils.reduce_utils import channelwise_minmax


class EMAObserver(AffineObserverBase):
    """
    Exponential-Moving-Average min/max tracker.

    Why?
    -----
    • Smoother than raw MinMax (reduces outlier shock).
    • Much cheaper than histogram/MSE observers.

    The update rule follows the common "momentum" form:

        ema = momentum * ema + (1 - momentum) * new_value

    With momentum → 0: FAST adaptation, momentum → 1: SLOW adaptation.
    """

    def __init__(
        self,
        *,
        momentum: float = 0.9,
        **kwargs,
    ):
        super().__init__(**kwargs)
        assert 0.0 < momentum < 1.0, "momentum must be in (0, 1)"
        self.momentum = momentum

    @torch.no_grad()
    def _update_stats(self, x: torch.Tensor):
        if self.channel_axis is None:
            curr_min, curr_max = x.min(), x.max()
        else:
            curr_min, curr_max = channelwise_minmax(x, self.channel_axis)

        if (
            torch.isinf(self.min_val).any() and torch.isinf(self.max_val).any()
        ):  # first batch → hard init
            self.min_val, self.max_val = curr_min, curr_max
            return

        m = self.momentum
        self.min_val = m * self.min_val + (1 - m) * curr_min
        self.max_val = m * self.max_val + (1 - m) * curr_max



# Copyright (c) 2025 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
from typing import Dict, Optional, Type, TypeVar

from tico.quantization.config.base import BaseConfig
from tico.quantization.quantizer import BaseQuantizer

TQ = TypeVar("TQ", bound=BaseQuantizer)

# Mapping: Config type -> Quantizer type
_REGISTRY: Dict[Type[BaseConfig], Type[BaseQuantizer]] = {}


def register_quantizer(config_cls: Type[BaseConfig]):
    """
    Decorator to register a quantizer for a given config class.
    Usage:
        @register_quantizer(GPTQConfig)
        class GPTQQuantizer(BaseQuantizer): ...
    """

    def wrapper(quantizer_cls: Type[TQ]) -> Type[TQ]:
        _REGISTRY[config_cls] = quantizer_cls
        return quantizer_cls

    return wrapper


def _lookup(cfg: BaseConfig) -> Optional[Type[BaseQuantizer]]:
    """Return a quantizer class only if the exact config type is registered."""
    return _REGISTRY.get(type(cfg))


def get_quantizer(cfg: BaseConfig) -> BaseQuantizer:
    """Factory to return a quantizer instance for the given config."""
    qcls = _lookup(cfg)
    if qcls is not None:
        return qcls(cfg)

    # Lazy import by naming convention
    name = getattr(cfg, "name", None)
    if name:
        if name == "ptq":
            importlib.import_module(f"tico.quantization.wrapq.quantizer")
        else:
            try:
                importlib.import_module(f"tico.quantization.algorithm.{name}.quantizer")
            except Exception as e:
                raise RuntimeError(
                    f"Failed to import quantizer module for config name='{name}': {e}"
                )

    qcls = _lookup(cfg)
    if qcls is not None:
        return qcls(cfg)

    raise RuntimeError(
        f"No quantizer registered for config type {type(cfg).__name__} "
        f"(name='{getattr(cfg,'name',None)}')."
    )



# Copyright (c) 2025 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import auto, Enum


class QScheme(Enum):
    # ───── Per-tensor ────────────
    PER_TENSOR_ASYMM = auto()
    PER_TENSOR_SYMM = auto()
    # ───── Per-channel ───────────
    PER_CHANNEL_ASYMM = auto()
    PER_CHANNEL_SYMM = auto()

    # helper
    def is_per_channel(self) -> bool:
        return self in {
            QScheme.PER_CHANNEL_ASYMM,
            QScheme.PER_CHANNEL_SYMM,
        }

    def is_symmetric(self) -> bool:
        return self in {
            QScheme.PER_TENSOR_SYMM,
            QScheme.PER_CHANNEL_SYMM,
        }

    def __str__(self) -> str:
        return self.name.lower()



# Copyright (c) 2026 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Optional, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F

from tico.quantization.config.ptq import PTQConfig
from tico.quantization.wrapq.mode import Mode
from tico.quantization.wrapq.observers.base import ObserverBase
from tico.quantization.wrapq.qscheme import QScheme
from tico.quantization.wrapq.wrappers.quant_module_base import QuantModuleBase
from tico.quantization.wrapq.wrappers.registry import register


@register(nn.Conv3d)
class QuantConv3dDecomposed(QuantModuleBase):
    """
    Quantization wrapper for nn.Conv3d with decomposition to Conv2d.

    This class decomposes Conv3d into multiple Conv2d operations to ensure
    all computations remain quantized. The decomposition follows the slice + Conv2d + Add
    approach, avoiding graph passes that introduce floating-point operations.

    Quantization:
    - Per-channel weight quantization (asymmetric)
    - Per-tensor input activation quantization
    - Per-tensor output activation quantization
    - Per-tensor quantization for all intermediate tensors (input slices, conv2d outputs, accumulators)
    """

    def __init__(
        self,
        fp: nn.Conv3d,
        *,
        qcfg: Optional[PTQConfig] = None,
        fp_name: Optional[str] = None,
    ):
        super().__init__(qcfg, fp_name=fp_name)

        # Static observers (always exist)
        self.obs_weight = self._make_obs(
            "weight", qscheme=QScheme.PER_CHANNEL_ASYMM, channel_axis=0
        )
        self.obs_act_in = self._make_obs("act_in")
        self.obs_act_out = self._make_obs("act_out")

        # Store original module
        self.module = fp

        # Dynamic observers (created lazily during first forward pass)
        self._input_slice_obs: Dict[int, ObserverBase] = {}  # Maps k (int) -> observer
        self._conv2d_obs: Dict[int, ObserverBase] = {}  # Maps k (int) -> observer
        self._acc_obs: Dict[int, ObserverBase] = {}  # Maps t_out (int) -> observer

        # Tracking for lazy observer creation
        self._dynamic_obs_calibrated = False

    def enable_calibration(self) -> None:
        """Enable calibration mode."""
        super().enable_calibration()

        # Collect weight statistics immediately (weights are static)
        self.obs_weight.collect(self.module.weight)

        # Reset dynamic observers for new calibration
        self._dynamic_obs_calibrated = False

    def _create_dynamic_observers(
        self,
        kT: int,
        T_out: int,
    ):
        """
        Create dynamic observers for intermediate quantization points.

        Args:
            kT: Kernel temporal dimension
            T_out: Number of output temporal positions
        """

        def create_observer(obs_name_prefix, obs_dictionary, dict_key):
            obs_name = f"{obs_name_prefix}{dict_key}"
            obs = self._make_obs(obs_name)
            obs_dictionary[dict_key] = obs
            self.add_module(obs_name, obs)
            # self.add_module(obs_name, obs) is required for torch.export() to properly access
            # the observer and its internal quantization parameters (cached_scale, cached_zp)
            # during graph construction. When torch.export() traces the model, it creates
            # 'get_attr' nodes to access module attributes. If observers are stored only in
            # dictionaries or via setattr(), torch.export() cannot create valid get_attr nodes
            # for the observer's cached_scale and cached_zp tensors, leading to warnings:
            #   "Attempted to insert a get_attr Node with no underlying reference"
            # By registering with add_module(), the observer becomes part of the module's
            # named_modules() tree, making both the observer AND its quantization parameters
            # accessible to the graph construction process. This ensures that the exported
            # graph can properly reference quantization parameters during Circle conversion.

        # Input slice observers (one for each temporal kernel position)
        for k in range(kT):
            create_observer(f"input_slice_k", self._input_slice_obs, k)

        # Conv2d output observers (one for each temporal kernel position)
        for k in range(kT):
            create_observer(f"conv2d_out_k", self._conv2d_obs, k)

        # Accumulator observers (one for each output temporal position)
        for t_out in range(T_out):
            create_observer(f"accumulator_t", self._acc_obs, t_out)

        self._dynamic_obs_calibrated = True

    def _parse_padding(self, padding) -> Tuple[int, int, int]:
        """Parse padding parameter to (temporal, height, width) tuple."""
        if isinstance(padding, str):
            if padding == "same":
                kT, kH, kW = self.module.kernel_size
                return kT // 2, kH // 2, kW // 2
            elif padding == "valid":
                return 0, 0, 0
            else:
                raise ValueError(f"Unsupported padding string: {padding}")
        elif isinstance(padding, (list, tuple)):
            if len(padding) == 1:
                return padding[0], padding[0], padding[0]
            elif len(padding) == 3:
                return padding[0], padding[1], padding[2]
            else:
                raise ValueError(f"Unsupported padding format: {padding}")
        elif isinstance(padding, int):  # int
            return padding, padding, padding
        else:
            raise ValueError(f"Unsupported padding type: {type(padding)}")

    def _apply_temporal_padding(
        self,
        x: torch.Tensor,
        temporal_padding: int,
    ) -> torch.Tensor:
        """Apply temporal padding using zeros and cat."""
        if temporal_padding == 0:
            return x

        N, C_in, T_in, H_in, W_in = x.shape

        # Create zero padding tensors
        zero_pad = torch.zeros(
            N, C_in, temporal_padding, H_in, W_in, dtype=x.dtype, device=x.device
        )

        # Cat: [zeros, input, zeros]
        padded = torch.cat([zero_pad, x, zero_pad], dim=2)

        return padded

    def _get_padded_input_slice(
        self,
        padded_x: torch.Tensor,
        t_idx: int,
        k: int,
    ) -> torch.Tensor:
        """
        Get and quantize input slice at temporal position.

        Args:
            padded_x: Temporally padded input tensor (N, C_in, T_padded, H_in, W_in)
            t_idx: Temporal index to slice
            k: Kernel temporal position (for observer lookup)

        Returns:
            Quantized input slice (N, C_in, 1, H_in, W_in)
        """
        # Slice at temporal position
        input_slice = padded_x[:, :, t_idx : t_idx + 1, :, :]

        # Quantize input slice
        input_slice_q = self._fq(input_slice, self._input_slice_obs[k])

        return input_slice_q

    def _apply_conv2d_quantized(
        self,
        input_2d: torch.Tensor,
        weight_slice: torch.Tensor,
        bias: Optional[torch.Tensor],
        k: int,
        H_out: int,
        W_out: int,
        padding: Tuple[int, int, int],
    ) -> torch.Tensor:
        """
        Apply quantized Conv2d operation.

        Args:
            input_2d: 2D input (N, C_in, H_in, W_in)
            weight_slice: 2D weight slice (C_out, C_in, kH, kW)
            bias: Optional bias tensor
            k: Kernel temporal position (for observer lookup)
            H_out: Output height
            W_out: Output width

        Returns:
            Quantized Conv2d output (N, C_out, H_out, W_out)
        """
        # Apply Conv2d
        conv_out = F.conv2d(
            input_2d,
            weight_slice,
            bias=None,  # Bias added after accumulation
            stride=(self.module.stride[1], self.module.stride[2]),
            padding=(padding[1], padding[2]),
            dilation=(self.module.dilation[1], self.module.dilation[2]),
            groups=self.module.groups,
        )

        # Quantize Conv2d output
        conv_out_q = self._fq(conv_out, self._conv2d_obs[k])

        return conv_out_q

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass with quantized Conv3d decomposition.

        Decomposes Conv3d into:
        1. Temporal padding (if needed)
        2. Slice input at each temporal kernel position
        3. Apply Conv2d to each slice
        4. Accumulate Conv2d results with quantization
        5. Add bias (if present)
        6. Stack temporal outputs

        Special case optimization:
        When kernel_size = input_size, stride = kernel_size,
        padding = 0, groups = 1, and dilation = 1 for all dimensions,
        the Conv3d operation reduces to matrix multiplication and is handled
        with a more efficient direct approach.

        All intermediate tensors are quantized to ensure integer-only computation.
        """
        N, C_in, T_in, H_in, W_in = x.shape
        C_out, C_in_weight, kT, kH, kW = self.module.weight.shape
        sT, sH, sW = self.module.stride
        dT, dH, dW = self.module.dilation
        groups = self.module.groups

        if C_in != C_in_weight:
            raise RuntimeError("Channels number mismatch")

        # Parse padding
        padding = self._parse_padding(self.module.padding)
        temporal_padding, h_padding, w_padding = padding

        # Quantize input activation
        x_q = self._fq(x, self.obs_act_in)

        # Get quantized weight
        w = self.module.weight
        if self._mode is Mode.QUANT:
            w = self.obs_weight.fake_quant(w)

        # Check for special case:
        # kernel_size = input_size,
        # stride = kernel_size,
        # padding = 0,
        # no dilation
        # groups = 1
        is_special_case = (
            (kT, kH, kW) == (T_in, H_in, W_in)
            and (sT, sH, sW) == (kT, kH, kW)
            and (temporal_padding, h_padding, w_padding) == (0, 0, 0)
            and (dT, dH, dW) == (1, 1, 1)
            and groups == 1
        )

        # Special case: Conv3d reduces to matrix multiplication
        if is_special_case:
            # Reshape input: (N, C_in, T_in, H_in, W_in) -> (N, 1, 1, C_in*T_in*H_in*W_in)
            x_q = x_q.reshape(N, 1, 1, -1)

            # Reshape weights: (C_out, C_in, kT, kH, kW) -> (C_out, 1, 1, C_in*kT*kH*kW)
            w = w.reshape(C_out, 1, 1, -1)

            # Apply Conv2d directly
            if self.module.bias is not None:
                conv2d_result = F.conv2d(x_q, w, self.module.bias)
            else:
                conv2d_result = F.conv2d(x_q, w)

            # Reshape output: (1, C_out, N*C_in, 1) -> (N, C_out, 1, 1, 1)
            result = conv2d_result.reshape(N, C_out, 1, 1, 1)

            # Quantize output activation
            result_q = self._fq(result, self.obs_act_out)
            return result_q

        # Normal case: Conv3d is decomposed to multiple Conv2D and Add operations
        else:
            # Calculate output dimensions
            T_padded = T_in + 2 * temporal_padding
            T_out = (T_padded - dT * (kT - 1) - 1) // sT + 1
            H_out = (H_in + 2 * h_padding - dH * (kH - 1) - 1) // sH + 1
            W_out = (W_in + 2 * w_padding - dW * (kW - 1) - 1) // sW + 1

            # Create dynamic observers on first forward pass
            if not self._dynamic_obs_calibrated:
                if self._mode is Mode.QUANT:
                    raise RuntimeError(
                        "Trying to quantize without calibration. Need to calibrate first."
                    )
                self._create_dynamic_observers(kT, T_out)

            # Apply temporal padding
            padded_input = self._apply_temporal_padding(x_q, temporal_padding)

            # Temporal processing loop
            temporal_outputs = []
            for t_out in range(T_out):
                t_in = t_out * sT
                accumulator = None

                for k in range(kT):
                    t_idx = t_in + k * dT

                    # Handle dilation: mask out-of-bounds positions
                    if dT > 1 and t_idx >= T_padded:
                        # Skip this kernel position (out of bounds)
                        continue

                    # Get and quantize input slice
                    input_slice_q = self._get_padded_input_slice(padded_input, t_idx, k)

                    # Remove temporal dimension: (N, C_in, 1, H_in, W_in) → (N, C_in, H_in, W_in)
                    input_2d = input_slice_q.squeeze(2)

                    # Slice weight at temporal position k
                    weight_slice = w[:, :, k, :, :]  # (C_out, C_in, kH, kW)

                    # Apply quantized Conv2d
                    conv_out_q = self._apply_conv2d_quantized(
                        input_2d,
                        weight_slice,
                        self.module.bias,
                        k,
                        H_out,
                        W_out,
                        padding,
                    )

                    # Accumulate with quantization
                    if accumulator is None:
                        accumulator = conv_out_q
                    else:
                        accumulator = self._fq(
                            accumulator + conv_out_q, self._acc_obs[t_out]
                        )

                # Add bias if present
                if self.module.bias is not None:
                    bias_reshaped = self.module.bias.reshape(1, C_out, 1, 1)
                    accumulator = accumulator + bias_reshaped

                temporal_outputs.append(accumulator)

            # Stack temporal outputs
            unsqueezed = [t.unsqueeze(2) for t in temporal_outputs]  # type: ignore[union-attr]
            stacked = torch.cat(unsqueezed, dim=2)  # (N, C_out, T_out, H_out, W_out)

            # Quantize output activation
            stacked_q = self._fq(stacked, self.obs_act_out)

            return stacked_q

    def _all_observers(self):
        """Return all observers for this module."""
        # Static observers
        yield from (self.obs_weight, self.obs_act_in, self.obs_act_out)

        # Dynamic observers (if created)
        yield from self._input_slice_obs.values()
        yield from self._conv2d_obs.values()
        yield from self._acc_obs.values()



# Copyright (c) 2025 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Type

from tico.config.base import CompileConfigBase
from tico.config.v1 import CompileConfigV1


class CompileConfigFactory:
    _config_classes = {
        "1.0": CompileConfigV1,
        # '2.0': CompileConfigV2,
    }

    @classmethod
    def get_config(cls, version: str) -> Type[CompileConfigBase]:
        if version not in cls._config_classes:
            raise ValueError(f"Unsupported version: {version}")

        return cls._config_classes[version]

    @classmethod
    def create(cls, version: str):
        config_class = cls.get_config(version)
        return config_class()


def get_default_config(version: str = "1.0"):
    return CompileConfigFactory.create(version)



# Copyright (c) 2025 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from typing import List, Optional, Tuple

import torch
import torch.nn as nn

from tico.quantization.config.ptq import PTQConfig
from tico.quantization.wrapq.wrappers.ptq_wrapper import PTQWrapper
from tico.quantization.wrapq.wrappers.quant_module_base import QuantModuleBase
from tico.quantization.wrapq.wrappers.registry import try_register


@try_register(
    "transformers.models.llama.modeling_llama.LlamaAttention",
    "transformers.models.llama.modeling_llama.LlamaSdpaAttention",
    variant="prefill",
)
class QuantLlamaAttentionPrefill(QuantModuleBase):
    def __init__(
        self,
        fp_attn: nn.Module,
        *,
        qcfg: Optional[PTQConfig] = None,
        fp_name: Optional[str] = None,
    ):
        super().__init__(qcfg, fp_name=fp_name)

        cfg = fp_attn.config
        self.config = cfg

        # head shapes
        assert hasattr(cfg, "hidden_size") and hasattr(cfg, "num_attention_heads")
        assert hasattr(cfg, "num_key_value_heads")
        assert isinstance(cfg.hidden_size, int) and isinstance(
            cfg.num_attention_heads, int
        )
        assert isinstance(cfg.num_key_value_heads, int)
        self.head_dim = getattr(
            cfg, "head_dim", cfg.hidden_size // cfg.num_attention_heads
        )
        self.kv_rep = cfg.num_attention_heads // cfg.num_key_value_heads
        self.n_kv = cfg.num_key_value_heads

        # ---- Wrap q k v o projections via PTQWrapper ---------------
        q_cfg = qcfg.child("q_proj") if qcfg else None
        k_cfg = qcfg.child("k_proj") if qcfg else None
        v_cfg = qcfg.child("v_proj") if qcfg else None
        o_cfg = qcfg.child("o_proj") if qcfg else None

        assert hasattr(fp_attn, "q_proj") and isinstance(
            fp_attn.q_proj, torch.nn.Module
        )
        assert hasattr(fp_attn, "k_proj") and isinstance(
            fp_attn.k_proj, torch.nn.Module
        )
        assert hasattr(fp_attn, "v_proj") and isinstance(
            fp_attn.v_proj, torch.nn.Module
        )
        assert hasattr(fp_attn, "o_proj") and isinstance(
            fp_attn.o_proj, torch.nn.Module
        )

        self.q_proj = PTQWrapper(
            fp_attn.q_proj, qcfg=q_cfg, fp_name=f"{fp_name}.q_proj"
        )
        self.k_proj = PTQWrapper(
            copy.deepcopy(fp_attn.k_proj), qcfg=k_cfg, fp_name=f"{fp_name}.k_proj"
        )
        self.v_proj = PTQWrapper(
            fp_attn.v_proj, qcfg=v_cfg, fp_name=f"{fp_name}.v_proj"
        )
        self.o_proj = PTQWrapper(
            fp_attn.o_proj, qcfg=o_cfg, fp_name=f"{fp_name}.o_proj"
        )

        # Constant scale (1/√d)
        scale_t = torch.tensor(
            float(getattr(fp_attn, "scaling", self.head_dim**-0.5))
        )
        # merge scale_t to k_proj, (otherwise merge it to q_proj)
        with torch.no_grad():
            lin = self.k_proj.wrapped.module
            lin.weight.mul_(scale_t)
            if lin.bias is not None:
                lin.bias.mul_(scale_t)

        mk = self._make_obs
        self.obs_hidden = mk("hidden")

        # RoPE tables
        self.obs_cos = mk("cos")
        self.obs_sin = mk("sin")

        # rotate_half sub-steps (q)
        self.obs_q_x1 = mk("q_x1")
        self.obs_q_x2 = mk("q_x2")
        self.obs_q_cat = mk("q_cat")

        # rotate_half sub-steps (k)
        self.obs_k_x1 = mk("k_x1")
        self.obs_k_x2 = mk("k_x2")
        self.obs_k_cat = mk("k_cat")

        # RoPE combine
        self.obs_q_cos = mk("q_cos")
        self.obs_q_sin = mk("q_sin")
        self.obs_q_rot = mk("q_rot")
        self.obs_k_cos = mk("k_cos")
        self.obs_k_sin = mk("k_sin")
        self.obs_k_rot = mk("k_rot")

        # Masking & attention math
        self.obs_causal_mask = mk("causal_mask")
        self.obs_logits = mk("logits")
        self.obs_mask_add = mk("mask_add")
        self.obs_softmax = mk("softmax")
        self.obs_attn_out = mk("attn_out")
        self.obs_attn_weights = mk("attn_weights")
        self.obs_attn_out_h = mk("attn_out_h")

        # Static causal mask template
        assert hasattr(cfg, "max_position_embeddings")
        max_seq = cfg.max_position_embeddings
        mask = torch.full((1, 1, max_seq, max_seq), float("-120"))  # type: ignore[arg-type]
        mask.triu_(1)
        self.register_buffer("causal_mask_template", mask, persistent=False)

    def _rot(self, t, o_x1, o_x2, o_cat):
        x1, x2 = torch.chunk(t, 2, dim=-1)
        x1 = self._fq(x1, o_x1)
        x2 = self._fq(x2, o_x2)
        x2n = x2
        return self._fq(torch.cat((x2n, x1), -1), o_cat)

    @staticmethod
    def _concat_kv(
        past: Optional[Tuple[torch.Tensor, torch.Tensor]],
        k_new: torch.Tensor,
        v_new: torch.Tensor,
        h_idx: int,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Concat along sequence dim (dim=1): (B, S, H) given head index."""
        if past is None:
            return k_new, v_new
        past_k, past_v = past
        k = torch.cat([past_k[:, h_idx, :, :], k_new], dim=1)
        v = torch.cat([past_v[:, h_idx, :, :], v_new], dim=1)
        return k, v

    def _apply_rope(
        self, t, cos, sin, obs_x1, obs_x2, obs_cat, obs_cos, obs_sin, obs_rot
    ):
        t_half = self._rot(t, obs_x1, obs_x2, obs_cat)
        t_cos = self._fq(t * cos, obs_cos)
        t_sin = self._fq(t_half * sin, obs_sin)
        t_rot = self._fq(t_cos + t_sin, obs_rot)

        return t_rot

    def forward(
        self,
        hidden_states: torch.Tensor,
        position_embeddings: tuple[torch.Tensor, torch.Tensor],
        attention_mask: Optional[torch.Tensor] = None,
        past_key_value=None,  # tuple(k, v) or HF Cache-like object
        use_cache: Optional[bool] = False,
        cache_position: Optional[torch.LongTensor] = None,
        **kwargs,
    ):
        hidden = self._fq(hidden_states, self.obs_hidden)
        B, S, _ = hidden.shape
        H = self.head_dim

        kv_rep = self.kv_rep
        q = self.q_proj(hidden).view(B, S, -1, H)  # (B, S, n_h, H)
        k = self.k_proj(hidden).view(B, S, -1, H)  # (B, K, n_kv, H)
        v = self.v_proj(hidden).view(B, S, -1, H)  # (B, K, n_kv, H)

        # Rope tables
        cos, sin = position_embeddings
        cos = self._fq(cos, self.obs_cos)
        sin = self._fq(sin, self.obs_sin)

        # --- KV for attention & present_key_value -------------
        present_key_value: Tuple[torch.Tensor, torch.Tensor]

        # Build causal mask if needed
        if attention_mask is None or attention_mask.dtype == torch.bool:
            q_len = q.size(1)
            k_len = k.size(1)
            assert isinstance(self.causal_mask_template, torch.Tensor)
            attention_mask = self.causal_mask_template[..., :q_len, :k_len].to(
                hidden_states.device
            )
            attention_mask = attention_mask.squeeze(0)
            attention_mask = self._fq(attention_mask, self.obs_causal_mask)

        attn_weights_parts = []
        attn_out_parts = []
        present_k_parts: List[torch.Tensor] = []
        present_v_parts: List[torch.Tensor] = []
        for kv_i in range(self.n_kv):
            # k_h, v_h: (B, K, H)
            k_i = k[:, :, kv_i, :]
            v_i = v[:, :, kv_i, :]

            k_i = self._apply_rope(
                k_i,
                cos,
                sin,
                self.obs_k_x1,
                self.obs_k_x2,
                self.obs_k_cat,
                self.obs_k_cos,
                self.obs_k_sin,
                self.obs_k_rot,
            )
            present_k_parts.append(k_i)
            present_v_parts.append(v_i)

            k_i, v_i = self._concat_kv(past_key_value, k_i, v_i, kv_i)
            for rep_i in range(kv_rep):
                q_idx = kv_i * kv_rep + rep_i
                # q_h: (B, S, H)
                q_i = q[:, :, q_idx, :]
                q_i = self._apply_rope(
                    q_i,
                    cos,
                    sin,
                    self.obs_q_x1,
                    self.obs_q_x2,
                    self.obs_q_cat,
                    self.obs_q_cos,
                    self.obs_q_sin,
                    self.obs_q_rot,
                )

                # logits: (B, S, K)
                logits_i = self._fq(q_i @ k_i.transpose(-2, -1), self.obs_logits)

                # mask add
                assert (
                    attention_mask.shape[-2:] == logits_i.shape[-2:]
                    and attention_mask.shape[0] == 1
                )  # check for compatiblity
                logits_i = self._fq(
                    logits_i + attention_mask,
                    self.obs_mask_add,
                )

                # softmax
                attn_i = torch.softmax(logits_i, -1, dtype=torch.float32).to(q_i.dtype)
                attn_i = self._fq(attn_i, self.obs_softmax)

                # out: (B, S, H)
                out_i = self._fq(attn_i @ v_i, self.obs_attn_out)

                attn_weights_parts.append(attn_i)
                attn_out_parts.append(out_i)

        # concat heads back
        # (B, n_h, S, K)
        attn_weights = self._fq(
            torch.stack(attn_weights_parts, dim=1), self.obs_attn_weights
        )
        # (B, n_h, S, H)
        attn_out_h = self._fq(torch.stack(attn_out_parts, dim=1), self.obs_attn_out_h)

        # Attention output: (B, S, n_h * H)
        attn_out = attn_out_h.transpose(1, 2).reshape(B, S, -1)

        # Final projection
        out = self.o_proj(attn_out)

        # Present KV: (B, n_kv, S, H)
        present_k = torch.stack(present_k_parts, dim=1)
        present_v = torch.stack(present_v_parts, dim=1)
        present_key_value = (present_k, present_v)

        # return with/without cache
        if use_cache:
            return out, attn_weights, present_key_value
        else:
            return out, attn_weights

    def _all_observers(self):
        # local first
        yield from (
            self.obs_hidden,
            self.obs_cos,
            self.obs_sin,
            self.obs_causal_mask,
            self.obs_q_x1,
            self.obs_q_x2,
            self.obs_q_cat,
            self.obs_k_x1,
            self.obs_k_x2,
            self.obs_k_cat,
            self.obs_q_cos,
            self.obs_q_sin,
            self.obs_q_rot,
            self.obs_k_cos,
            self.obs_k_sin,
            self.obs_k_rot,
            self.obs_logits,
            self.obs_mask_add,
            self.obs_softmax,
            self.obs_attn_out,
            self.obs_attn_weights,
            self.obs_attn_out_h,
        )
        # recurse into children that are QuantModuleBase
        for m in (self.q_proj, self.k_proj, self.v_proj, self.o_proj):
            yield from m._all_observers()



# Copyright (c) 2026 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Optional

import torch
import torch.nn as nn

from transformers.cache_utils import Cache
from transformers.modeling_outputs import CausalLMOutputWithPast

from tico.quantization.config.ptq import PTQConfig
from tico.quantization.wrapq.wrappers.ptq_wrapper import PTQWrapper
from tico.quantization.wrapq.wrappers.quant_module_base import QuantModuleBase
from tico.quantization.wrapq.wrappers.registry import try_register


@try_register("transformers.models.llama.modeling_llama.LlamaForCausalLM")
class QuantLlamaForCausalLM(QuantModuleBase):
    def __init__(
        self,
        model_fp: nn.Module,
        *,
        qcfg: Optional[PTQConfig] = None,
        fp_name: Optional[str] = None,
    ):
        super().__init__(qcfg, fp_name=fp_name)
        self.quantizers = getattr(model_fp, "quantizers", None)

        # ----- child configs (hierarchical override) -------------------
        model_cfg = qcfg.child("model") if qcfg else None
        lm_head_cfg = qcfg.child("lm_head") if qcfg else None

        ## ----- wrap model/lm_head -------------------------------
        assert hasattr(model_fp, "model") and isinstance(
            model_fp.model, torch.nn.Module
        )
        assert hasattr(model_fp, "lm_head") and isinstance(
            model_fp.lm_head, torch.nn.Module
        )

        self.model = PTQWrapper(
            model_fp.model, qcfg=model_cfg, fp_name=f"{fp_name}.model"
        )

        self.lm_head = PTQWrapper(
            model_fp.lm_head, qcfg=lm_head_cfg, fp_name=f"{fp_name}.lm_head"
        )
        self.config = model_fp.config
        self.loss_function = model_fp.loss_function
        self.device = model_fp.device

    def tie_weights(self):
        pass

    def forward(
        self,
        input_ids: torch.LongTensor | None = None,
        attention_mask: torch.Tensor | None = None,
        position_ids: torch.LongTensor | None = None,
        past_key_values: Cache | None = None,
        inputs_embeds: torch.FloatTensor | None = None,
        labels: torch.LongTensor | None = None,
        use_cache: bool | None = None,
        cache_position: torch.LongTensor | None = None,
        logits_to_keep: int | torch.Tensor = 0,
        **kwargs,
    ) -> CausalLMOutputWithPast:

        output_attentions = self.config.output_attentions
        output_hidden_states = self.config.output_hidden_states
        return_dict = self.config.use_return_dict

        # decoder outputs consists of (dec_features, layer_state, dec_hidden, dec_attn)
        outputs = self.model(
            input_ids=input_ids,
            attention_mask=attention_mask,
            position_ids=position_ids,
            past_key_values=past_key_values,
            inputs_embeds=inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
            cache_position=cache_position,
            **kwargs,
        )

        hidden_states = outputs[0]
        # Only compute necessary logits, and do not upcast them to float if we are not computing the loss
        slice_indices = (
            slice(-logits_to_keep, None)
            if isinstance(logits_to_keep, int)
            else logits_to_keep
        )
        logits = self.lm_head(hidden_states[:, slice_indices, :])

        loss = None
        if labels is not None:
            loss = self.loss_function(
                logits=logits,
                labels=labels,
                vocab_size=self.config.vocab_size,
                **kwargs,
            )

        if not return_dict:
            output = (logits,) + outputs[1:]
            return (loss,) + output if loss is not None else output

        return CausalLMOutputWithPast(
            loss=loss,
            logits=logits,
            past_key_values=outputs.past_key_values,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )

    def _all_observers(self):
        # recurse into children that are QuantModuleBase
        for m in (self.model, self.lm_head):
            yield from m._all_observers()



# Copyright (c) 2025 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import torch.fx
import torch
from torch.export import ExportedProgram

from tico.serialize.circle_mapping import extract_shape
from tico.utils.graph import add_placeholder, create_node
from tico.utils.passes import PassBase, PassResult
from tico.utils.trace_decorators import trace_graph_diff_on_pass
from tico.utils.utils import is_target_node, set_new_meta_val
from tico.utils.validate_args_kwargs import AddmmArgs


@trace_graph_diff_on_pass
class DecomposeAddmm(PassBase):
    """
    Let's decompose addmm to add + mul + matmul.

    [BEFORE]

    input   mat1    mat2    beta    alpha
    |       |       |       |       |
    --------------addmm--------------
                    |
                    out

    [AFTER]

    input   beta    mat1    mat2    alpha
    |       |       |       |        |
    ---mul---       ---mm----        |
        |               |            |
        |               -----mul-----
        |                     |
        ---------add----------
                  |
                  out

    """

    def __init__(self):
        super().__init__()

    def call(self, exported_program: ExportedProgram) -> PassResult:
        gm = exported_program.graph_module
        graph: torch.fx.Graph = gm.graph
        modified = False

        for node in graph.nodes:
            if not is_target_node(node, torch.ops.aten.addmm.default):
                continue

            args = AddmmArgs(*node.args, **node.kwargs)
            input = args.input
            mat1 = args.mat1
            mat2 = args.mat2
            beta = args.beta
            alpha = args.alpha

            with graph.inserting_before(node):
                # out = beta * input + alpha * (mat1 @ mat2)
                matmul = create_node(
                    graph, torch.ops.aten.mm.default, (mat1, mat2), origin=node
                )
                set_new_meta_val(matmul)

                if beta == 1:
                    bias: torch.fx.Node | torch.Tensor = input
                elif beta == 0:
                    bias = add_placeholder(
                        exported_program,
                        torch.zeros(extract_shape(input)),
                        f"{node.name}_beta_zeros",
                    )
                else:
                    bias = create_node(
                        graph, torch.ops.aten.mul.Tensor, (input, beta), origin=node
                    )

                if alpha == 1:
                    scaled_matmul: torch.fx.Node | torch.Tensor = matmul
                elif alpha == 0:
                    scaled_matmul = add_placeholder(
                        exported_program,
                        torch.zeros(extract_shape(matmul)),
                        f"{node.name}_alpha_zeros",
                    )
                else:
                    scaled_matmul = create_node(
                        graph, torch.ops.aten.mul.Tensor, (matmul, alpha), origin=node
                    )

                result = create_node(
                    graph, torch.ops.aten.add.Tensor, (bias, scaled_matmul)
                )

            node.replace_all_uses_with(result, propagate_meta=True)

            modified = True

        gm.graph.eliminate_dead_code()
        gm.graph.lint()
        gm.recompile()

        return PassResult(modified)



# Copyright IST-DASLab. 2025. (commit: 2d65066). GitHub repository.
# Retrieved from https://github.com/IST-DASLab/gptq. Licensed under the
# Apache License 2.0.

# Copyright (c) 2025 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# https://github.com/IST-DASLab/gptq/blob/2d65066/gptq.py

import math
import time
from typing import Optional

import torch
import torch.nn as nn

from tico.quantization.algorithm.gptq.gptq import (
    conv2d_weights_to_convtranspose2d_weights,
    convtranspose2d_weights_to_conv2d_weights,
    get_matmul_input_for_convtranspose2d,
)

from tico.quantization.algorithm.gptq.quant import quantize, Quantizer


def iterate_GPTQ(scale, zero, maxq, W, Hinv, max_num_of_iters=50):

    cur_weights = W.clone()
    mults = torch.pow(torch.diag(Hinv), -1)
    Hinv_U = torch.triu(Hinv, diagonal=1)

    init_weights = W.clone()
    for _ in range(max_num_of_iters):
        cur_Q = quantize(cur_weights, scale, zero, maxq)

        d_W = torch.mul((cur_weights - cur_Q), mults)
        cur_weights = init_weights - torch.matmul(d_W, Hinv_U)
        del d_W, cur_Q
        d_W = cur_Q = None

    del init_weights
    init_weights = None

    cur_Q = quantize(cur_weights, scale, zero, maxq)

    return cur_Q, cur_weights


class FPI_GPTQ:
    def __init__(self, layer):
        self.layer = layer
        self.dev = self.layer.weight.device
        W = layer.weight.data.clone()
        if isinstance(self.layer, nn.Conv2d) or isinstance(self.layer, nn.Conv1d):
            W = W.flatten(1)
        elif isinstance(self.layer, nn.ConvTranspose2d):
            W = convtranspose2d_weights_to_conv2d_weights(self.layer, W)
            W = W.flatten(1)

        self.rows = W.shape[0]
        self.columns = W.shape[1]
        self.H: Optional[torch.Tensor] = torch.zeros(
            (self.columns, self.columns), device=self.dev
        )
        self.nsamples = 0
        self.quantizer: Quantizer = Quantizer()

    def add_batch(self, inp, out):
        if len(inp.shape) == 2:
            inp = inp.unsqueeze(0)
        tmp = inp.shape[0]
        if isinstance(self.layer, nn.Linear):
            if len(inp.shape) > 2:
                inp = inp.reshape((-1, inp.shape[-1]))
            inp = inp.t()
        if isinstance(self.layer, nn.Conv2d):
            unfold = nn.Unfold(
                self.layer.kernel_size,
                dilation=self.layer.dilation,
                padding=self.layer.padding,
                stride=self.layer.stride,
            )

            if self.layer.groups != 1:
                # the idea behind conversion of depthwise convolution to matmul is described here
                # https://discuss.pytorch.org/t/conv1d-implementation-using-torch-nn-functional-unfold/109643/2
                # although depthwise convolution is equal to a set of MatMuls
                # (please note `w.view(1, groups, out_channels // groups, -1)` in the reference above is not just w.flatten(1))
                # we can approximate groupwise Hessians with their mean
                # so that we will have just a single Hessian and the usual GPTQ applies
                inp = inp.reshape(
                    inp.size(0) * self.layer.groups,
                    inp.size(1) // self.layer.groups,
                    inp.shape[2],
                    inp.shape[3],
                )  # inp.shape == (batch*groups, in_channels / groups, H, W) to meet Groupwise-wise Convolution, so that each group is colvolved with its own filter

            inp = unfold(
                inp
            )  # inp.shape == (batch*groups, k_h*k_w*in_channels / groups, flattened_patches)
            inp = inp.permute(
                [1, 0, 2]
            )  # inp.shape == (k_h*k_w*in_channels / groups, batch * groups, flattened_patches)
            inp = inp.flatten(
                1
            )  # inp.shape == (k_h*k_w*in_channels / groups, batch * groups * flattened_patches)
            # so inp.matmul(inp.t()).shape == (k_x*k_y*in_channels / groups, k_x*k_y*in_channels / groups) == W.flatten(1)

        if isinstance(self.layer, nn.Conv1d):
            # nn.Conv1d is basically the same as nn.Conv2d so we can use the same idea as for nn.Conv2d
            # TODO reduce code duplication
            # represent conv1d as conv2d(1, k) on reshaped_input(batch, in_channels, 1, L)
            unfold = nn.Unfold(
                (1, self.layer.kernel_size[0]),
                dilation=(1, self.layer.dilation[0]),
                padding=(0, self.layer.padding[0]),
                stride=(1, self.layer.stride[0]),
            )
            if self.layer.groups != 1:
                # please see Conv2D for additional info
                inp = inp.reshape(
                    inp.size(0) * self.layer.groups,
                    inp.size(1) // self.layer.groups,
                    inp.shape[2],
                )  # inp.shape == (batch*groups, in_channels / groups, L) to meet Groupwise-wise Convolution, so that each group is colvolved with its own filter

            inp = inp.unsqueeze(
                -2
            )  # (batch*groups, in_channels / groups, L)->(batch*groups, in_channels / groups, 1, L), valid for Conv2D
            inp = unfold(inp)
            inp = inp.permute([1, 0, 2])
            inp = inp.flatten(1)
        if isinstance(self.layer, nn.ConvTranspose2d):
            inp = get_matmul_input_for_convtranspose2d(self.layer, inp)

        self.H *= self.nsamples / (self.nsamples + tmp)
        self.nsamples += tmp
        inp = math.sqrt(2 / self.nsamples) * inp.float()
        self.H += inp.matmul(inp.t())

    def fasterquant(
        self,
        percdamp=0.01,
        verbose=False,
    ):
        W = self.layer.weight.data.clone()
        if isinstance(self.layer, nn.Conv2d) or isinstance(self.layer, nn.Conv1d):
            W = W.flatten(1)
        elif isinstance(self.layer, nn.ConvTranspose2d):
            W = convtranspose2d_weights_to_conv2d_weights(self.layer, W)
            conv2d_shape = W.shape
            W = W.flatten(1)  # reshaped to matrix (OUT_channels x the_rest)

        W = W.float()
        tick = time.time()
        if not self.quantizer.ready():
            self.quantizer.find_params(W, weight=True)

        H = self.H
        del self.H
        assert isinstance(H, torch.Tensor)
        dead = torch.diag(H) == 0
        H[dead, dead] = 1
        W[:, dead] = 0

        # actorder
        perm = torch.argsort(torch.diag(H), descending=True)
        W = W[:, perm]
        H = H[perm][:, perm]
        invperm = torch.argsort(perm)

        Q = torch.zeros_like(W)

        damp = percdamp * torch.mean(torch.diag(H))
        diag = torch.arange(self.columns, device=self.dev)
        H[diag, diag] += damp
        H = torch.linalg.cholesky(H)
        assert isinstance(H, torch.Tensor)
        H = torch.cholesky_inverse(H)
        H = torch.linalg.cholesky(H, upper=True)
        Hinv = H

        Q, W = iterate_GPTQ(
            self.quantizer.scale,
            self.quantizer.zero,
            self.quantizer.maxq,
            W,
            Hinv=Hinv,
            max_num_of_iters=min(
                50, self.columns
            ),  # we don't need to iterate more than self.columns
        )

        if torch.cuda.is_available():
            torch.cuda.synchronize()
        if verbose:
            print("time %.2f" % (time.time() - tick))
            Losses = 0.5 * ((Q - W) / torch.diag(Hinv)) ** 2
            print("error", torch.sum(Losses).item())

        Q = Q[:, invperm]

        if isinstance(self.layer, nn.Conv2d) or isinstance(self.layer, nn.Conv1d):
            Q[:, dead] = quantize(
                self.layer.weight.flatten(1)[:, dead],
                self.quantizer.scale,
                self.quantizer.zero,
                self.quantizer.maxq,
            )
        elif isinstance(self.layer, nn.ConvTranspose2d):
            Q[:, dead] = quantize(
                convtranspose2d_weights_to_conv2d_weights(
                    self.layer, self.layer.weight.data
                ).flatten(1)[:, dead],
                self.quantizer.scale,
                self.quantizer.zero,
                self.quantizer.maxq,
            )
        else:
            Q[:, dead] = quantize(
                self.layer.weight[:, dead],
                self.quantizer.scale,
                self.quantizer.zero,
                self.quantizer.maxq,
            )

        if isinstance(self.layer, nn.ConvTranspose2d):
            Q_conv2d = Q.reshape(conv2d_shape).to(self.layer.weight.data.dtype)
            self.layer.weight.data = conv2d_weights_to_convtranspose2d_weights(
                self.layer, Q_conv2d
            )
        else:
            self.layer.weight.data = Q.reshape(self.layer.weight.shape).to(
                self.layer.weight.data.dtype
            )

    def free(self):
        self.H = None
        if torch.cuda.is_available():
            torch.cuda.empty_cache()



# Copyright (c) 2025 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from torch.export import ExportedProgram

from tico.passes import ops
from tico.utils import logging
from tico.utils.passes import PassBase, PassResult
from tico.utils.trace_decorators import trace_graph_diff_on_pass
from tico.utils.utils import is_target_node
from tico.utils.validate_args_kwargs import CatArgs


@trace_graph_diff_on_pass
class MergeConsecutiveCat(PassBase):
    """
    This pass merges consecutive `aten.cat` operators when they can be merged into single operator.
    """

    def __init__(self):
        super().__init__()

    def call(self, exported_program: ExportedProgram) -> PassResult:
        logger = logging.getLogger(__name__)

        graph_module = exported_program.graph_module
        graph = graph_module.graph
        modified = False
        for cat in graph.nodes:
            if not is_target_node(cat, ops.aten.cat):
                continue

            args = CatArgs(*cat.args, **cat.kwargs)  # type: ignore[arg-type]
            inputs = args.tensors
            dim = args.dim

            new_inputs = []
            for prev_cat in inputs:
                new_inputs.append(prev_cat)
                if not prev_cat.op == "call_function":
                    continue

                if prev_cat.target not in ops.aten.cat:
                    continue

                prev_args = CatArgs(*prev_cat.args, **prev_cat.kwargs)  # type: ignore[arg-type]
                prev_inputs = prev_args.tensors
                prev_dim = prev_args.dim

                if not prev_dim == dim:
                    continue

                new_inputs.pop()
                for prev_input in prev_inputs:
                    new_inputs.append(prev_input)

            if len(new_inputs) > len(inputs):
                cat.args = (new_inputs, dim)

                modified = True
                logger.debug(
                    f"Consecutive cat nodes before {cat.name} are merged into {cat.name}"
                )

        graph.eliminate_dead_code()
        graph.lint()
        graph_module.recompile()

        return PassResult(modified)



# Copyright (c) 2025 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import List, Optional

import torch

"""
This is a key for torch.fx.Node's meta dict to save QuantParam

QuantParam can be retrieved as node.meta[QPARAM_KEY]
"""
QPARAM_KEY = "_quantization_parameters_"


@dataclass
class QuantParam:
    scale: Optional[List[float]] = None
    zero_point: Optional[List[int]] = None
    quantized_dimension: Optional[int] = None
    min: Optional[List[float]] = None
    max: Optional[List[float]] = None
    # NOTE We define dtype as a string to easily extend new dtypes (ex: uint4)
    dtype: str = ""


def to_qparam_dtype(dtype: torch.dtype) -> str:
    str_type = str(dtype)
    assert str_type.startswith("torch.")
    return str_type[6:]



# Copyright (c) 2026 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from typing import Iterable, Optional

import torch
import torch.nn as nn

from tico.quantization.config.ptq import PTQConfig
from tico.quantization.wrapq.wrappers.ptq_wrapper import PTQWrapper
from tico.quantization.wrapq.wrappers.quant_module_base import QuantModuleBase
from tico.quantization.wrapq.wrappers.registry import try_register


@try_register(
    "transformers.models.qwen3_vl.modeling_qwen3_vl.Qwen3VLVisionAttention",
)
class QuantQwen3VLVisionAttention(QuantModuleBase):
    def __init__(
        self,
        attn_fp: nn.Module,
        *,
        qcfg: Optional[PTQConfig] = None,
        fp_name: Optional[str] = None,
    ):
        super().__init__(qcfg, fp_name=fp_name)
        cfg = attn_fp.config
        self.config = cfg
        self.num_heads = attn_fp.num_heads
        self.head_dim = attn_fp.head_dim

        # ---- Wrap q k v o projections via PTQWrapper ---------------
        qkv_cfg = qcfg.child("qkv") if qcfg else None
        proj_cfg = qcfg.child("proj") if qcfg else None

        assert hasattr(attn_fp, "qkv") and isinstance(attn_fp.qkv, torch.nn.Module)
        assert hasattr(attn_fp, "proj") and isinstance(attn_fp.proj, torch.nn.Module)

        self.qkv = PTQWrapper(
            copy.deepcopy(attn_fp.qkv), qcfg=qkv_cfg, fp_name=f"{fp_name}.qkv_cfg"
        )
        self.proj = PTQWrapper(attn_fp.proj, qcfg=proj_cfg, fp_name=f"{fp_name}.proj")

        # Let's fold constant scale (1/√d) to k_proj
        scale_t = torch.tensor(
            float(getattr(attn_fp, "scaling", self.head_dim**-0.5))
        )
        with torch.no_grad():
            lin = self.qkv.wrapped.module
            k_offset = lin.weight.shape[0] // 3
            k_size = lin.weight.shape[0] // 3
            lin.weight[k_offset : k_offset + k_size, :].mul_(scale_t)
            if lin.bias is not None:
                lin.bias[k_offset : k_offset + k_size].mul_(scale_t)

        mk = self._make_obs
        self.obs_hidden = mk("hidden")
        self.obs_scaling = mk("scaling")
        self.obs_mul_logits_scale = mk("mul_logits_scale")
        self.obs_cos = mk("cos")
        self.obs_sin = mk("sin")

        # rotate_half sub-steps (q)
        self.obs_q_x1 = mk("q_x1")
        self.obs_q_x2 = mk("q_x2")
        self.obs_q_neg = mk("q_neg")
        self.obs_q_cat = mk("q_cat")

        # rotate_half sub-steps (k)
        self.obs_k_x1 = mk("k_x1")
        self.obs_k_x2 = mk("k_x2")
        self.obs_k_neg = mk("k_neg")
        self.obs_k_cat = mk("k_cat")

        # RoPE combine
        self.obs_q_cos = mk("q_cos")
        self.obs_q_sin = mk("q_sin")
        self.obs_q_rot = mk("q_rot")
        self.obs_k_cos = mk("k_cos")
        self.obs_k_sin = mk("k_sin")
        self.obs_k_rot = mk("k_rot")

        # Masking & attention math
        self.obs_logits = mk("logits")
        self.obs_softmax = mk("softmax")
        self.obs_attn_out = mk("attn_out")

    def _rot(self, t, o_x1, o_x2, o_neg, o_cat):
        x1, x2 = torch.chunk(t, 2, dim=-1)
        x1 = self._fq(x1, o_x1)
        x2 = self._fq(x2, o_x2)
        x2n = self._fq(-x2, o_neg)
        return self._fq(torch.cat((x2n, x1), -1), o_cat)

    def _apply_rope(
        self, q: torch.Tensor, k: torch.Tensor, cos: torch.Tensor, sin: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor]:
        cos, sin = cos.unsqueeze(-2), sin.unsqueeze(-2)
        q_rot = self._rot(
            q, self.obs_q_x1, self.obs_q_x2, self.obs_q_neg, self.obs_q_cat
        )
        q_cos = self._fq(q * cos, self.obs_q_cos)
        q_sin = self._fq(q_rot * sin, self.obs_q_sin)
        q_embed = self._fq(q_cos + q_sin, self.obs_q_rot)

        k_rot = self._rot(
            k, self.obs_k_x1, self.obs_k_x2, self.obs_k_neg, self.obs_k_cat
        )
        k_cos = self._fq(k * cos, self.obs_k_cos)
        k_sin = self._fq(k_rot * sin, self.obs_k_sin)
        k_embed = self._fq(k_cos + k_sin, self.obs_k_rot)

        return q_embed, k_embed

    def forward(
        self,
        hidden_states: torch.Tensor,
        cu_seqlens: torch.Tensor,
        rotary_pos_emb: torch.Tensor | None = None,
        position_embeddings: tuple[torch.Tensor, torch.Tensor] | None = None,
        **kwargs,
    ):
        hidden_states = self._fq(hidden_states, self.obs_hidden)

        seq_length = hidden_states.shape[0]
        query_states, key_states, value_states = (
            self.qkv(hidden_states)
            .reshape(seq_length, 3, self.num_heads, -1)
            .permute(1, 0, 2, 3)
            .unbind(0)
        )
        cos, sin = position_embeddings  # type: ignore[misc]
        cos = self._fq(cos, self.obs_cos)
        sin = self._fq(sin, self.obs_sin)
        query_states, key_states = self._apply_rope(query_states, key_states, cos, sin)
        query_states = query_states.transpose(0, 1).unsqueeze(0)
        key_states = key_states.transpose(0, 1).unsqueeze(0)
        value_states = value_states.transpose(0, 1).unsqueeze(0)

        # Other implementations: Process each chunk separately
        lengths = cu_seqlens[1:] - cu_seqlens[:-1]
        splits = [
            torch.split(tensor, lengths.tolist(), dim=2)
            for tensor in (query_states, key_states, value_states)
        ]
        assert (
            len(splits)
            == 3  # just a single image, so (q, k, v) = (splits[0], splits[1], splits[2])
            and len(splits[0]) == 1
            and len(splits[1]) == 1
            and len(splits[2]) == 1
        )  # so we may proceed without splitted attention
        rep = query_states.size(-3) // key_states.size(-3)
        assert rep == 1  # currently no GQA is supported

        attn_outputs = []
        for n_head in range(self.num_heads):
            k_i = key_states[:, n_head : n_head + 1, :, :]
            v_i = value_states[:, n_head : n_head + 1, :, :]
            q_i = query_states[:, n_head : n_head + 1, :, :]
            logits_i = self._fq(q_i @ k_i.transpose(-2, -1), self.obs_logits)
            # softmax
            attn_i = torch.softmax(logits_i, -1, dtype=torch.float32).to(q_i.dtype)
            attn_i = self._fq(attn_i, self.obs_softmax)
            out_i = self._fq(attn_i @ v_i, self.obs_attn_out)
            attn_outputs.append(out_i)

        attn_output = torch.cat(attn_outputs, dim=1)
        attn_output = attn_output.transpose(1, 2).contiguous()
        attn_output = attn_output.reshape(seq_length, -1).contiguous()

        attn_output = self.proj(attn_output)
        return attn_output

    def _all_observers(self) -> Iterable:
        yield from (
            self.obs_hidden,
            self.obs_scaling,
            self.obs_mul_logits_scale,
            self.obs_cos,
            self.obs_sin,
            self.obs_q_x1,
            self.obs_q_x2,
            self.obs_q_neg,
            self.obs_q_cat,
            self.obs_k_x1,
            self.obs_k_x2,
            self.obs_k_neg,
            self.obs_k_cat,
            self.obs_q_cos,
            self.obs_q_sin,
            self.obs_q_rot,
            self.obs_k_cos,
            self.obs_k_sin,
            self.obs_k_rot,
            self.obs_logits,
            self.obs_softmax,
            self.obs_attn_out,
        )
        for m in (self.qkv, self.proj):
            yield from m._all_observers()



import importlib.metadata
import importlib.util

from packaging import version

MIN_VERSIONS = {
    "llama": "4.36.0",  # transformers==4.31.0 supports llama but without layer_idx and position_embeddings feature
    "qwen3-vl": "4.57.0",
}


def has_transformers_for(model_type: str) -> bool:
    if importlib.util.find_spec("transformers") is None:
        return False

    current_version = importlib.metadata.version("transformers")
    required_version = MIN_VERSIONS.get(model_type)

    if not required_version:
        raise ValueError(f"Invalid model_type {model_type}")

    if version.parse(current_version) >= version.parse(required_version):
        return True

    return False



# Copyright (c) 2025 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib

import torch
from transformers import AutoModelForCausalLM, AutoTokenizer

from tico.quantization import convert, prepare
from tico.quantization.config.ptq import PTQConfig
from tico.quantization.evaluation.metric import compute_peir
from tico.quantization.evaluation.utils import plot_two_outputs
from tico.quantization.wrapq.mode import Mode
from tico.quantization.wrapq.wrappers.llama.quant_attn_prefill import (
    QuantLlamaAttentionPrefill,
)
from tico.utils.utils import SuppressWarning

name = "Maykeye/TinyLLama-v0"
MAX_SEQ = 256
model = AutoModelForCausalLM.from_pretrained(name, dtype=torch.float32)
tokenizer = AutoTokenizer.from_pretrained(name)

# Make sure pad token exists (Llama often uses eos as pad)
if tokenizer.pad_token_id is None:
    tokenizer.pad_token = tokenizer.eos_token

model.config.max_position_embeddings = MAX_SEQ

# -------------------------------------------------------------------------
# 1. Replace layer-0’s MLP with QuantLlamaMLP
# -------------------------------------------------------------------------
orig_attn = model.model.layers[0].self_attn
model.model.layers[0].self_attn = prepare(
    orig_attn, PTQConfig(wrapper_variant="prefill")
)
model.eval()

attn_q = model.model.layers[0].self_attn  # quant wrapper
assert isinstance(attn_q.wrapped, QuantLlamaAttentionPrefill)
rotary = model.model.rotary_emb

# -------------------------------------------------------------------------
# Helpers: tokenize + embed
# -------------------------------------------------------------------------
def make_inputs(prompt: str):
    batch = tokenizer(
        prompt,
        return_tensors="pt",
        padding="max_length",
        truncation=True,
        max_length=MAX_SEQ,
    )
    input_ids = batch["input_ids"]

    hidden = model.model.embed_tokens(input_ids)
    return hidden


# -------------------------------------------------------------------------
# 2. Single-pass calibration
# -------------------------------------------------------------------------
PROMPTS = [
    "The quick brown fox jumps over the lazy dog.",
    "In 2025, AI systems accelerated hardware-software co-design at scale.",
    "양자화는 왜 어려울까? 분포, 길이, 마스크가 관건이다.",
    "今日はいい天気ですね。ところでRoPE角度は長さに依存します。",
    "def quicksort(arr):\n    if len(arr) <= 1: return arr\n    ...",
    "Prices rose 3.14% — see Figure 2; emails: foo@bar.com!",
]

with torch.no_grad():
    for prompt in PROMPTS:
        embeds = make_inputs(prompt)
        position_ids = torch.arange(embeds.size(1)).unsqueeze(0)
        pos = rotary(embeds, position_ids)
        _ = attn_q(embeds, pos)  # observers collect

convert(attn_q)

assert attn_q._mode is Mode.QUANT, "Quantization mode should be active now."

# -------------------------------------------------------------------------
# 3. Quick diff check (INT-sim vs FP32)
# -------------------------------------------------------------------------
emb = make_inputs("check")
position_ids = torch.arange(emb.size(1)).unsqueeze(0)
pos = rotary(emb, position_ids)
with torch.no_grad():
    int8 = attn_q(emb, pos)[0]
    fp32 = orig_attn(emb, position_embeddings=pos, attention_mask=None)[0]

print("┌───────────── Quantization Error Summary ─────────────")
print(f"│ Mean |diff|: {(int8 - fp32).abs().mean().item():.6f}")
print(f"│ PEIR       : {compute_peir(fp32, int8) * 100:.6f} %")
print("└──────────────────────────────────────────────────────")
print(plot_two_outputs(fp32, int8))

# -------------------------------------------------------------------------
# 4. Export the quantized block
# -------------------------------------------------------------------------
import tico

save_path = pathlib.Path("attn_prefill.q.circle")
B, S, D = 1, MAX_SEQ, model.config.hidden_size
example = torch.randn(B, S, D)
example_pos = rotary(example, torch.arange(S)[None, :])

with SuppressWarning(UserWarning, ".*"):
    cm = tico.convert(attn_q, (example, example_pos))
cm.save(save_path)

print(f"Quantized Circle model saved to {save_path.resolve()}")



# Copyright (c) 2025 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import torch.fx
import torch
from torch._export.utils import is_lifted_tensor_constant
from torch.export import ExportedProgram

from tico.passes import ops
from tico.utils import logging
from tico.utils.passes import PassBase, PassResult
from tico.utils.trace_decorators import trace_graph_diff_on_pass
from tico.utils.utils import is_target_node
from tico.utils.validate_args_kwargs import AddTensorArgs


@trace_graph_diff_on_pass
class LegalizeCausalMaskValue(PassBase):
    """
    This pass replaces occurrences of -inf in attention masks with a large negative finite value (e.g., -120) to ensure numerical stability in computations, particularly in softmax operations.

    This pass can be turned enable only when
        1. The model will be quantized later (e.g., by circle-quantizer).
        2. Softmax kernel of our backend does not support masking.
        3. `Add with -inf` is used only for masking.
    """

    def __init__(self, enabled: bool = False):
        super().__init__()
        self.enabled = enabled

    def call(self, exported_program: ExportedProgram) -> PassResult:
        if not self.enabled:
            return PassResult(False)

        new_mask = -120  # Make it configurable
        logger = logging.getLogger(__name__)

        graph_module = exported_program.graph_module
        graph = graph_module.graph
        modified = False

        for node in graph.nodes:
            if not is_target_node(node, ops.aten.add):
                continue

            args = AddTensorArgs(*node.args, **node.kwargs)
            input = args.input
            other = args.other

            if isinstance(input, torch.fx.Node) and is_lifted_tensor_constant(
                exported_program, input
            ):
                mask_node = input
            elif isinstance(other, torch.fx.Node) and is_lifted_tensor_constant(
                exported_program, other
            ):
                mask_node = other
            else:
                continue

            assert isinstance(mask_node, torch.fx.Node)

            mask_node_name = (
                exported_program.graph_signature.inputs_to_lifted_tensor_constants[
                    mask_node.name
                ]
            )
            mask_data = exported_program.constants[mask_node_name]

            # WHY Use -1.e+38, not -float('inf') or torch.finfo(torch.float32).min?
            #
            # torch.finfo(torch.float32).min is -3.4028234663852886e+38 but it changes while processed in const prop or other passes.
            # Therefore, use a rounded value and compare to know it's very large negative number.
            fp32_minus_inf_rounded = -1.0e38
            if torch.all(
                torch.logical_or(mask_data == 0, mask_data < fp32_minus_inf_rounded)
            ):
                # Replace the value from -inf to -120
                exported_program.constants[mask_node_name] = torch.where(
                    mask_data < fp32_minus_inf_rounded,
                    torch.tensor(new_mask, dtype=mask_data.dtype),
                    mask_data,
                )

            modified = False  # To run only once
            logger.debug(
                f"{mask_node.name}'s mask data are changed from '-inf' to {new_mask}"
            )

        graph.eliminate_dead_code()
        graph.lint()
        graph_module.recompile()

        return PassResult(modified)



# Copyright (c) 2024 Intel Corporation
# Copyright (c) 2025 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import tqdm


def find_layers(module, layers=[torch.nn.Linear], name=""):
    if type(module) in layers:
        return {name: module}
    res = {}
    for name1, child in module.named_children():
        res.update(
            find_layers(
                child, layers=layers, name=name + "." + name1 if name != "" else name1
            )
        )
    return res


def gather_single_batch_from_dict(data_dict, idx):
    """
    Gather single batch from a dict.

    Args:
        data_dict (dict): data dict.
        idx (int): index

    Returns:
        dict: single batch.
    """
    # obtain a set of keyword input from cache
    single_batch = {}
    for k, v in data_dict.items():
        single_batch[k] = data_dict[k][idx]
    return single_batch


def gather_single_batch_from_list(data_list, idx):
    """
    Gather single batch from a list.

    Args:
        data_dict (dict): data list.
        idx (int): index

    Returns:
        list: single batch.
    """
    # obtain a set of positional input from cache
    single_batch = []
    for data_item in data_list:
        single_batch.append(data_item[idx])
    return single_batch


def get_dataset_for_calibration(model, dataset):
    """Enrich dataset with model ouputs to be used as targets"""

    class DataSetWithLabels(torch.utils.data.Dataset):
        def __init__(self, inputs, targets, transform=None):
            self.n_inputs = len(inputs)
            self.inputs = inputs
            self.labels = targets
            self.transform = transform

        def __len__(self):
            return self.n_inputs

        def __getitem__(self, idx):
            if torch.is_tensor(idx):
                idx = idx.tolist()
            sample = self.inputs[idx]
            if self.transform:
                sample = self.transform(sample)

            return (sample, self.labels[idx])

    targets = []

    with torch.no_grad():
        print("Computing calibration set")
        for prompt in tqdm.tqdm(dataset):
            results = model(prompt.to(model.device)).logits.detach()
            results = torch.argmax(results.detach(), dim=-1).cpu()

            targets.append(results)

    labeled_data = DataSetWithLabels(dataset, targets)
    dataloader = torch.utils.data.DataLoader(labeled_data, batch_size=1, shuffle=False)
    return dataloader


class SensitivityCalibrator:
    """
    Sensitivity calibrator - compute sensitivies using empirical Fisher information.

    Sensitivities are assumed to mimick second order derivatives.
    They can be used to estimate `logits` global error introduced by quantization.
    So here we use empirical Fisher information to assess diagonal second order derivatives.
    Please see https://arxiv.org/abs/1905.12558?ref=inference.vc for a discussion.
    """

    def __init__(self, model, tokenizer, dataset):
        self.model = model
        self.tokenizer = tokenizer
        self.dataset = dataset

    def compute_sensitivity_info(self):

        data_loader = get_dataset_for_calibration(self.model, self.dataset)

        dtype = self.model.dtype
        model = self.model.float()

        sensitivity = {}
        modules_to_process = {}
        name_of_module: dict[torch.nn.Linear, str] = {}

        for name, module in model.named_modules():
            if isinstance(module, torch.nn.Linear):
                modules_to_process[name] = module
                name_of_module[module] = name
                sensitivity[name] = torch.zeros_like(module.weight).cpu()

        print("Calibrating sensitivity")
        for inputs, targets in tqdm.tqdm(data_loader):
            model.zero_grad()
            inp_ids = inputs.view(-1, inputs.shape[-1])
            logits = model(inp_ids.to(model.device)).logits

            outputs = logits.squeeze()
            targets = targets.squeeze()

            t_index = outputs.shape[0] - 1  # priority to the last token
            outputs_el = outputs[t_index : t_index + 1, :]  # noqa E203
            targets_el = targets[t_index : t_index + 1]  # noqa E203

            model.zero_grad()
            loss = torch.nn.CrossEntropyLoss()(
                outputs_el, targets_el.to(model.device)
            )  # for Fisher this must be CrossEntropy

            loss.backward(retain_graph=False)

            # update second order information as current weights gradients are ready
            for name in modules_to_process:
                cur_module = modules_to_process[name]
                cur_grad = cur_module.weight.grad.detach().clone()
                if torch.isnan(cur_grad).any().item():
                    print("WARNING NaN detected")

                sensitivity[name] += torch.mul(cur_grad, cur_grad).cpu()

                cur_grad = None
                del cur_grad

                if model.device.type != "cpu":
                    torch.cuda.empty_cache()
                    torch.cuda.synchronize()

            loss.detach()

            logits = outputs = targets = loss = None
            del loss, logits, outputs, targets

            if model.device.type != "cpu":
                torch.cuda.empty_cache()
                torch.cuda.synchronize()

        for name in modules_to_process:
            sensitivity[name] /= len(data_loader)

        model = model.to(dtype)

        return sensitivity



# Copyright (c) 2025 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ============================================================================
# LAYER-WISE DIFF DEBUGGING PIPELINE
# ----------------------------------------------------------------------------
# A quantization debugging pipeline that identifies accuracy regressions
# by comparing UINT vs FP outputs at each layer.
#
#   1. Load a full-precision (FP) LLaMA-3-1B model.
#   2. Wrap each Transformer block with PTQWrapper (activations → fake-quant).
#   3. Capture reference FP layer outputs before quantization.
#   4. Calibrate UINT-8 activation observers in a single pass.
#   5. Freeze quantization parameters (scale, zero-point).
#   6. Re-run inference and compare UINT-8 vs FP outputs per layer.
#   7. Report where quantization hurts the most.
#
# Use this pipeline to trace precision loss layer by layer, and pinpoint
# problematic modules during post-training quantization.
# ============================================================================

import argparse
import sys

import torch
import tqdm
from datasets import load_dataset
from transformers import AutoModelForCausalLM, AutoTokenizer

from tico.quantization import convert, prepare
from tico.quantization.config.ptq import PTQConfig
from tico.quantization.wrapq.utils.introspection import (
    build_fqn_map,
    compare_layer_outputs,
    save_fp_outputs,
)
from tico.quantization.wrapq.wrappers.ptq_wrapper import PTQWrapper

# Token-budget presets for activation calibration
TOKENS: dict[str, int] = {
    # Smoke test (<1 min turnaround on CPU/GPU)
    "debug": 2_000,  # ≈16 × 128-seq batches
    # Good default for 1-7B models (≲3 % ppl delta)
    "baseline": 50_000,
    # Production / 4-bit observer smoothing
    "production": 200_000,
}

DTYPE_MAP = {
    "float32": torch.float32,
    # TODO Support more dtypes
    # "bfloat16": torch.bfloat16,
    # "float16": torch.float16,
}

# Hardcoded dataset settings
DATASET_NAME = "wikitext"
DATASET_CONFIG = "wikitext-2-raw-v1"
TRAIN_SPLIT = "train"


def main():
    parser = argparse.ArgumentParser(
        description="Layer-wise diff debugging pipeline for PTQ"
    )
    parser.add_argument(
        "--model", type=str, required=True, help="HF repo name or local path."
    )
    parser.add_argument(
        "--device",
        type=str,
        default="cuda" if torch.cuda.is_available() else "cpu",
        help="Device to run on (cuda|cpu|mps).",
    )
    parser.add_argument(
        "--dtype",
        choices=list(DTYPE_MAP.keys()),
        default="float32",
        help=f"Model dtype for load.",
    )
    parser.add_argument(
        "--stride",
        type=int,
        default=512,
        help="Sliding-window stride used during calibration.",
    )
    parser.add_argument(
        "--calib-preset",
        choices=list(TOKENS.keys()),
        default="debug",
        help="Calibration token budget preset.",
    )
    parser.add_argument("--seed", type=int, default=42, help="Random seed.")
    parser.add_argument(
        "--trust-remote-code",
        action="store_true",
        help="Enable only if you trust the model repo code.",
    )
    parser.add_argument(
        "--hf-token",
        type=str,
        default=None,
        help="Optional HF token for gated/private repos.",
    )
    parser.add_argument(
        "--use-cache",
        dest="use_cache",
        action="store_true",
        default=False,
        help="Use model KV cache if enabled (off by default).",
    )
    parser.add_argument(
        "--no-tqdm", action="store_true", help="Disable tqdm progress bars."
    )

    args = parser.parse_args()

    # Basic setup
    torch.manual_seed(args.seed)
    device = torch.device(args.device)
    dtype = DTYPE_MAP[args.dtype]  # noqa: E999 (kept readable)

    print("=== Config ===")
    print(f"Model            : {args.model}")
    print(f"Device           : {device.type}")
    print(f"DType            : {args.dtype}")
    print(f"Stride           : {args.stride}")
    print(
        f"Calib preset     : {args.calib_preset} ({TOKENS[args.calib_preset]:,} tokens)"
    )
    print(f"Use HF cache?    : {args.use_cache}")
    print()

    # -------------------------------------------------------------------------
    # 1. Load the FP backbone and tokenizer
    # -------------------------------------------------------------------------
    print("Loading FP model …")
    tokenizer = AutoTokenizer.from_pretrained(
        args.model,
        trust_remote_code=args.trust_remote_code,
        token=args.hf_token,
        legacy=False,
    )
    model = (
        AutoModelForCausalLM.from_pretrained(
            args.model,
            dtype=dtype,
            trust_remote_code=args.trust_remote_code,
            token=args.hf_token,
        )
        .to(device)
        .eval()
    )

    # Disable KV cache to force full forward passes for introspection
    model.config.use_cache = args.use_cache

    # Build module -> FQN map before wrapping
    m_to_fqn = build_fqn_map(model)

    # Prepare calibration inputs (HF Wikitext-2 train split)
    CALIB_TOKENS = TOKENS[args.calib_preset]
    print(f"Calibrating with {CALIB_TOKENS:,} tokens.\n")
    # Use Wikitext-2 train split for calibration.
    dataset = load_dataset(DATASET_NAME, DATASET_CONFIG, split=TRAIN_SPLIT)

    # -------------------------------------------------------------------------
    # 2. Wrap every layer with PTQWrapper (UINT-8 activations)
    # -------------------------------------------------------------------------
    print("Wrapping layers with PTQWrapper …")
    qcfg = PTQConfig(wrapper_variant="prefill")  # default: per-tensor UINT8
    model = prepare(model, qcfg)

    # -------------------------------------------------------------------------
    # 3. Activation calibration plus FP-vs-UINT8 diffing
    # -------------------------------------------------------------------------
    print("Calibrating UINT-8 observers …")
    calib_txt = " ".join(dataset["text"])[:CALIB_TOKENS]
    ids = tokenizer(calib_txt, return_tensors="pt").input_ids.to(device)

    # Save reference FP activations before observers clamp/quantize
    save_handles, act_cache = save_fp_outputs(model)

    iterator = range(0, ids.size(1) - 1, args.stride)
    if not args.no_tqdm:
        iterator = tqdm.tqdm(iterator, desc="Act-Calibration")
    with torch.no_grad():
        for i in iterator:
            inputs = ids[:, i : i + args.stride]
            model(inputs)  # observers collect act. ranges

    # Remove save hooks now that FP activations are cached
    for h in save_handles:
        h.remove()

    # Freeze (scale, zero-point) after calibration
    convert(model)

    # Register diff hooks and measure per-layer deltas
    cmp_handles = compare_layer_outputs(model, act_cache, metrics=["diff", "peir"])
    # Use same inputs for comparison.
    with torch.no_grad():
        model(inputs)

    assert isinstance(cmp_handles, list)
    for h in cmp_handles:
        h.remove()


if __name__ == "__main__":
    try:
        main()
    except Exception as e:
        print(f"\n[Error] {e}", file=sys.stderr)
        sys.exit(1)



# Copyright (c) 2025 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import torch
import tqdm


def _resolve_device(device: torch.device, model: torch.nn.Module):
    if device != "auto":
        return torch.device(device)

    for p in model.parameters():
        if p is not None:
            return p.device

    # fallback
    return torch.device("cpu")


def perplexity(
    model: torch.nn.Module,
    encodings: torch.Tensor,
    device: torch.device | str = "auto",
    *,
    max_length: Optional[int] = None,
    stride: int = 512,
    ignore_index: int | None = -100,
    show_progress: bool = True,
) -> float:
    """
    Compute perplexity (PPL) using a "strided sliding-window"
     evaluation strategy.

    The function:
    1. Splits the token sequence into overlapping windows of length
       `max_length` (model context size).
    2. Masks tokens that were already scored in previous windows
       (`labels == -100`), so each token's negative log-likelihood (NLL)
       is counted EXACTLY once.
    3. Aggregates token-wise NLL to return corpus-level PPL.

    Parameters
    ----------
    model : torch.nn.Module
        Causal LM loaded in evaluation mode (`model.eval()`).
    encodings : torch.Tensor | transformers.BatchEncoding
        Tokenised corpus.  If a `BatchEncoding` is passed, its
        `.input_ids` field is used.  Shape must be `(1, seq_len)`.
    device : torch.device | str
        CUDA or CPU device on which to run evaluation.
    max_length : int, optional
        Context window size.  Defaults to `model.config.max_position_embeddings`.
    stride : int, default = 512
        Step size by which the sliding window advances.  Must satisfy
        `1 ≤ stride ≤ max_length`.
    ignore_index : int, default = -100
        Label value to ignore in loss computation. This should match
        the `ignore_index` used by the model's internal
        `CrossEntropyLoss`. For Hugging Face causal LMs, the
        convention is `-100`.
    show_progress : bool, default = True
        If True, displays a tqdm progess bar while evaluating.

    Returns
    -------
    float
        Corpus-level perplexity.
    """
    # -------- input preparation -------- #
    try:
        # transformers.BatchEncoding has `input_ids`
        input_ids_full = encodings.input_ids  # type: ignore[attr-defined]
    except AttributeError:  # already a tensor
        input_ids_full = encodings
    assert isinstance(input_ids_full, torch.Tensor)
    device = _resolve_device(device, model)
    input_ids_full = input_ids_full.to(device)

    if max_length is None:
        if hasattr(model, "config"):
            assert hasattr(model, "config")
            model_config = model.config
        else:
            assert hasattr(model.wrapped, "config")
            model_config = model.wrapped.config

        if hasattr(model_config, "text_config"):
            model_config = model_config.text_config
        assert hasattr(model_config, "max_position_embeddings")
        assert isinstance(model_config.max_position_embeddings, int)
        max_length = model_config.max_position_embeddings
    assert max_length is not None
    assert (
        1 <= stride <= max_length
    ), f"stride ({stride}) must be in [1, max_length ({max_length})]"

    seq_len = input_ids_full.size(1)
    nll_sum = 0.0
    n_tokens = 0
    prev_end = 0

    # -------- main loop -------- #
    for begin in tqdm.trange(0, seq_len, stride, desc="PPL", disable=not show_progress):
        end = min(begin + max_length, seq_len)
        trg_len = end - prev_end  # fresh tokens in this window

        input_ids = input_ids_full[:, begin:end]
        target_ids = input_ids.clone()
        # mask previously-scored tokens
        target_ids[:, :-trg_len] = ignore_index  # type: ignore[assignment]

        with torch.no_grad():
            outputs = model(input_ids, labels=target_ids)
            # loss is already averaged over non-masked labels
            neg_log_likelihood = outputs.loss

        # exact number of labels that contributed to loss
        loss_tokens = (target_ids[:, 1:] != ignore_index).sum().item()  # type: ignore[attr-defined]
        nll_sum += neg_log_likelihood * loss_tokens
        n_tokens += int(loss_tokens)

        prev_end = end
        if end == seq_len:
            break

    avg_nll: float | torch.Tensor = nll_sum / n_tokens
    if not isinstance(avg_nll, torch.Tensor):
        avg_nll = torch.tensor(avg_nll)
    assert isinstance(avg_nll, torch.Tensor)
    ppl = torch.exp(avg_nll)

    return ppl.item()



# Copyright (c) 2025 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# =============================================================================
# PTQ + GPTQ HYBRID QUANTIZATION PIPELINE
# -----------------------------------------------------------------------------
# This script shows how to:
#   1. Load a pretrained FP Llama-3 model.
#   2. Run GPTQ to quantize weights only.
#   3. Wrap every Transformer layer with a PTQWrapper to quantize activations.
#   4. Calibrate UINT-8 observers in a single pass over a text corpus.
#   5. Inject GPTQ’s per-tensor weight scales / zero-points into the PTQ graph.
#   6. Freeze all Q-params and compute Wikitext-2 perplexity.
# =============================================================================

import argparse
import sys
from typing import Any

import torch
import tqdm
from datasets import load_dataset
from transformers import AutoModelForCausalLM, AutoTokenizer

from tico.quantization import convert, prepare
from tico.quantization.config.gptq import GPTQConfig
from tico.quantization.config.ptq import PTQConfig
from tico.quantization.wrapq.observers.affine_base import AffineObserverBase
from tico.quantization.wrapq.utils.introspection import build_fqn_map
from tico.quantization.wrapq.utils.metrics import perplexity
from tico.quantization.wrapq.wrappers.ptq_wrapper import PTQWrapper
from tico.quantization.wrapq.wrappers.quant_module_base import QuantModuleBase

# Token-budget presets for activation calibration
TOKENS: dict[str, int] = {
    # Smoke test (<1 min turnaround on CPU/GPU)
    "debug": 2_000,  # ≈16 × 128-seq batches
    # Good default for 1-7B models (≲3 % ppl delta)
    "baseline": 50_000,
    # Production / 4-bit observer smoothing
    "production": 200_000,
}

DTYPE_MAP = {
    "float32": torch.float32,
    # TODO Support more dtypes
    # "bfloat16": torch.bfloat16,
    # "float16": torch.float16,
}

# Hardcoded dataset settings
DATASET_NAME = "wikitext"
DATASET_CONFIG = "wikitext-2-raw-v1"
TRAIN_SPLIT = "train"
TEST_SPLIT = "test"


# -------------------------------------------------------------------------
# 1. Helper — copy GPTQ (scale, zp) into PTQ observers
# -------------------------------------------------------------------------
def inject_gptq_qparams(
    root: torch.nn.Module,
    gptq_quantizers: dict[str, Any],  # {fp_name: quantizer}
    weight_obs_name: str = "weight",
):
    """
    For every `QuantModuleBase` whose `fp_name` matches a GPTQ key,
    locate the observer called `weight_obs_name` and overwrite its
    (scale, zero-point), then lock them against further updates.
    """
    for m in root.modules():
        if not isinstance(m, QuantModuleBase):
            continue
        if m.fp_name is None:
            continue
        quantizer = gptq_quantizers.get(m.fp_name)
        if quantizer is None:
            continue
        obs = m.get_observer(weight_obs_name)
        if obs is None:
            continue
        assert isinstance(obs, AffineObserverBase)
        # GPTQ quantizer attributes
        obs.load_qparams(quantizer.scale, quantizer.zero, lock=True)


def main():
    parser = argparse.ArgumentParser(
        description="GPTQ+PTQ pipeline (weight-only + activation UINT8)"
    )
    parser.add_argument(
        "--model", type=str, required=True, help="HF repo name or local path."
    )
    parser.add_argument(
        "--device",
        type=str,
        default="cuda" if torch.cuda.is_available() else "cpu",
        help="Device to run on (cuda|cpu|mps).",
    )
    parser.add_argument(
        "--dtype",
        choices=list(DTYPE_MAP.keys()),
        default="float32",
        help="Model dtype for load.",
    )
    parser.add_argument(
        "--stride",
        type=int,
        default=512,
        help="Sliding-window stride used for calibration and eval.",
    )
    parser.add_argument(
        "--max-seq-len",
        type=int,
        default=None,
        help="Cap the max sequence length fed to the model (calib + eval). "
        "If not set, uses full available length.",
    )
    parser.add_argument(
        "--calib-preset",
        choices=list(TOKENS.keys()),
        default="debug",
        help="Activation calibration token budget preset.",
    )
    parser.add_argument("--seed", type=int, default=42, help="Random seed.")
    parser.add_argument(
        "--trust-remote-code",
        action="store_true",
        help="Enable only if you trust the model repo code.",
    )
    parser.add_argument(
        "--hf-token",
        type=str,
        default=None,
        help="Optional HF token for gated/private repos.",
    )
    parser.add_argument(
        "--use-cache",
        dest="use_cache",
        action="store_true",
        default=False,
        help="Use model KV cache if enabled (off by default).",
    )
    parser.add_argument(
        "--no-tqdm", action="store_true", help="Disable tqdm progress bars."
    )

    args = parser.parse_args()

    # Basic setup
    torch.manual_seed(args.seed)
    device = torch.device(args.device)
    dtype = DTYPE_MAP[args.dtype]

    print("=== Config ===")
    print(f"Model            : {args.model}")
    print(f"Device           : {device.type}")
    print(f"DType            : {args.dtype}")
    print(f"Stride           : {args.stride}")
    print(f"Max seq len cap  : {args.max_seq_len}")
    print(
        f"Calib preset     : {args.calib_preset} ({TOKENS[args.calib_preset]:,} tokens)"
    )
    print(f"Use HF cache?    : {args.use_cache}")
    print()

    # -------------------------------------------------------------------------
    # 2. Load the FP backbone and tokenizer
    # -------------------------------------------------------------------------
    print("Loading FP model …")
    tokenizer = AutoTokenizer.from_pretrained(
        args.model,
        trust_remote_code=args.trust_remote_code,
        token=args.hf_token,
        legacy=False,
    )
    model = (
        AutoModelForCausalLM.from_pretrained(
            args.model,
            dtype=dtype,
            trust_remote_code=args.trust_remote_code,
            token=args.hf_token,
        )
        .to(device)
        .eval()
    )

    model.config.use_cache = args.use_cache

    # Build module -> FQN map BEFORE wrapping
    m_to_fqn = build_fqn_map(model)

    # -------------------------------------------------------------------------
    # 3. Run GPTQ (weight-only) pass
    # -------------------------------------------------------------------------
    print("Applying GPTQ …")
    dataset_test = load_dataset(DATASET_NAME, DATASET_CONFIG, split=TEST_SPLIT)
    q_m = prepare(model, GPTQConfig(), inplace=True)

    it = (
        dataset_test
        if args.no_tqdm
        else tqdm.tqdm(dataset_test, desc="GPTQ calibration")
    )
    for d in it:
        ids = tokenizer(d["text"], return_tensors="pt").input_ids.to(device)
        q_m(ids)  # observers gather weight stats

    q_m = convert(q_m, inplace=True)  # materialize INT-weight tensors

    # -------------------------------------------------------------------------
    # 4. Wrap every layer with PTQWrapper (activation UINT-8)
    # -------------------------------------------------------------------------
    print("Wrapping layers with PTQWrapper …")
    qcfg = PTQConfig(wrapper_variant="prefill")  # default: per-tensor UINT8
    q_m = prepare(q_m, qcfg)

    # -------------------------------------------------------------------------
    # 5. Single-pass activation calibration
    # -------------------------------------------------------------------------
    print("Calibrating UINT-8 observers …")
    CALIB_TOKENS = TOKENS[args.calib_preset]
    print(f"Calibrating with {CALIB_TOKENS:,} tokens.\n")
    dataset_train = load_dataset(DATASET_NAME, DATASET_CONFIG, split=TRAIN_SPLIT)
    calib_txt = " ".join(dataset_train["text"])[:CALIB_TOKENS]
    train_ids = tokenizer(calib_txt, return_tensors="pt").input_ids.to(device)

    # Overwrite weight observers with GPTQ statistics
    if hasattr(q_m, "quantizers") and isinstance(q_m.quantizers, dict):
        inject_gptq_qparams(q_m, q_m.quantizers)
    else:
        print(
            "[Warn] q_m.quantizers not found or not a dict; skipping GPTQ qparam injection."
        )

    # Forward passes to collect activation ranges
    iterator = range(0, train_ids.size(1) - 1, args.stride)
    if not args.no_tqdm:
        iterator = tqdm.tqdm(iterator, desc="Act-calibration")
    with torch.no_grad():
        for i in iterator:
            q_m(train_ids[:, i : i + args.stride])

    # Freeze all Q-params (scale, zero-point)
    convert(q_m)

    # -------------------------------------------------------------------------
    # 6. Evaluate perplexity on Wikitext-2
    # -------------------------------------------------------------------------
    print("\nCalculating perplexities …")
    enc = tokenizer("\n\n".join(dataset_test["text"]), return_tensors="pt")
    ppl_uint8 = perplexity(
        q_m, enc, device, max_length=args.max_seq_len, stride=args.stride
    )

    print("\n┌── Wikitext-2 test perplexity ─────────────")
    print(f"│ UINT-8 : {ppl_uint8:8.2f}")
    print("└───────────────────────────────────────────")


if __name__ == "__main__":
    try:
        main()
    except Exception as e:
        print(f"\n[Error] {e}", file=sys.stderr)
        sys.exit(1)



# Copyright (c) 2026 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib

import torch
from transformers import AutoModelForImageTextToText, AutoTokenizer

from tico.quantization import convert, prepare
from tico.quantization.config.ptq import PTQConfig
from tico.quantization.evaluation.metric import compute_peir
from tico.quantization.evaluation.utils import plot_two_outputs
from tico.quantization.wrapq.mode import Mode
from tico.quantization.wrapq.wrappers.qwen_vl.quant_text_mlp import QuantQwen3VLTextMLP
from tico.utils.utils import SuppressWarning

torch.manual_seed(123)


# -------------------------------------------------------------------------
# 0. Load a Qwen3-VL model (text tower) + tokenizer
# -------------------------------------------------------------------------
name = "Qwen/Qwen3-VL-2B-Instruct"
model = AutoModelForImageTextToText.from_pretrained(
    name,
    device_map="cpu",
    trust_remote_code=True,
)
tokenizer = AutoTokenizer.from_pretrained(name, trust_remote_code=True)

# Make sure pad token exists (Llama often uses eos as pad)
if tokenizer.pad_token_id is None:
    tokenizer.pad_token = tokenizer.eos_token

"""
As max_seq increases, the proportion of padded tokens grows, 
 which directly affects calibration statistics and tends to increase PEIR.

This is not a modeling issue but a calibration artifact: 
 observers see a distribution dominated by padding rather than real tokens.

For a more representative and realistic accuracy evaluation, 
the calibration dataset should be adjusted (e.g., longer or more diverse 
 sequence lengths, or padding-aware calibration) so that activation 
statistics better reflect actual inference workloads.
"""
MAX_SEQ = 128
text_cfg = model.config.text_config
text_cfg.max_position_embeddings = MAX_SEQ

# -------------------------------------------------------------------------
# 1. Replace layer-0's MLP with QuantQwen3VLTextMLP
# -------------------------------------------------------------------------
orig_mlp = model.model.language_model.layers[0].mlp
mlp_q = prepare(orig_mlp, PTQConfig())
mlp_q.eval()

assert isinstance(mlp_q.wrapped, QuantQwen3VLTextMLP)

# -------------------------------------------------------------------------
# 2. Single-pass calibration
# -------------------------------------------------------------------------
CALIB_TENSORS = []
for _ in range(10):
    ct = torch.randn(5, MAX_SEQ, text_cfg.hidden_size)
    CALIB_TENSORS.append(ct)

with torch.no_grad():
    for ct in CALIB_TENSORS:
        _ = mlp_q(ct)

convert(mlp_q)
assert mlp_q._mode is Mode.QUANT, "Quantization mode should be active now."

# -------------------------------------------------------------------------
# 3. Quick diff check (INT-sim vs FP32)
# -------------------------------------------------------------------------
hidden = CALIB_TENSORS[3]

with torch.no_grad():
    quant_out = mlp_q(hidden)
    fp_out = orig_mlp(hidden)

print("┌───────────── Quantization Error Summary ─────────────")
print(f"│ Mean |diff|: {(quant_out - fp_out).abs().mean().item():.6f}")
print(f"│ PEIR       : {compute_peir(fp_out, quant_out) * 100:.6f} %")
print("└──────────────────────────────────────────────────────")
print(plot_two_outputs(fp_out, quant_out))

# -------------------------------------------------------------------------
# 4. Export the quantized block
# -------------------------------------------------------------------------
import tico

save_path = pathlib.Path("qwen3vl_text_mlp.q.circle")
B, S, D = 1, MAX_SEQ, text_cfg.hidden_size
example = torch.randn(B, S, D)

with SuppressWarning(UserWarning, ".*"):
    cm = tico.convert(mlp_q, (example,))
cm.save(save_path)

print(f"Quantized Circle model saved to {save_path.resolve()}")



# Copyright (c) 2025 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, TYPE_CHECKING

if TYPE_CHECKING:
    import torch._ops
    import torch.fx
import torch
import torch.ao.quantization.fx._decomposed  # register `quantize_per_tensor`
from circle_schema import circle

from tico.serialize.circle_graph import CircleSubgraph
from tico.serialize.operators.hashable_opcode import OpCode
from tico.serialize.operators.node_visitor import NodeVisitor, register_node_visitor
from tico.serialize.operators.utils import create_builtin_operator, get_op_index
from tico.utils.validate_args_kwargs import QuantizePerTensorArgs


@register_node_visitor
class QuantizePerTensorDefaultVisitor(NodeVisitor):
    target: List[torch._ops.OpOverload] = [
        torch.ops.quantized_decomposed.quantize_per_tensor.default
    ]

    def __init__(self, op_codes: Dict[OpCode, int], graph: CircleSubgraph):
        super().__init__(op_codes, graph)

    def define_node(
        self,
        node: torch.fx.Node,
    ) -> circle.Operator.OperatorT:
        args = QuantizePerTensorArgs(*node.args, **node.kwargs)  # type: ignore[arg-type]
        tensor = args.tensor
        scale = args.scale
        zero_p = args.zero_p
        quant_min = args.quant_min
        quant_max = args.quant_max

        output_tensor: circle.Tensor.TensorT = self.graph.get_tensor(node)
        assert output_tensor.quantization is not None

        # Tensor should have qparam when it's exported
        # The qparam must match with the arguments of this Op
        assert output_tensor.quantization.scale[0] == scale
        assert output_tensor.quantization.zeroPoint[0] == zero_p

        if output_tensor.type == circle.TensorType.TensorType.UINT8:
            assert quant_min == 0 and quant_max == 255
        elif output_tensor.type == circle.TensorType.TensorType.INT16:
            # Some frameworks use -32767 as quant_min of int16
            assert quant_min in (-32768, -32767) and quant_max == 32767

        inputs = [tensor]
        outputs = [node]

        op_index = get_op_index(
            circle.BuiltinOperator.BuiltinOperator.QUANTIZE, self._op_codes
        )
        operator = create_builtin_operator(self.graph, op_index, inputs, outputs)

        # Op-specific option
        operator.builtinOptionsType = (
            circle.BuiltinOptions.BuiltinOptions.QuantizeOptions
        )
        option = circle.QuantizeOptions.QuantizeOptionsT()
        operator.builtinOptions = option

        return operator



# Copyright (c) 2025 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Tuple

import torch
import torch.fx._pytree as fx_pytree
import torch.utils._pytree as pytree

from tico.utils import logging
from tico.utils.installed_packages import is_transformers_installed

__all__ = [
    "register_dynamic_cache",
    "register_static_cache",
    "register_dynamic_layer",
    "register_static_layer",
    "register_encoder_decoder_cache",
]

##################################################################################
# [Dynamic/Static Cache Registration]
# Reference: https://github.com/Samsung/TICO/issues/417
#
# {transformer-version}: {status (implementation detail)}
# 4.51.0: ✅ Pre-registered (list-based)
# 4.51.1: ✅ Pre-registered (list-based)
# 4.52.0: ✅ Pre-registered (list-based)
# 4.52.1: ✅ Pre-registered (list-based)
# 4.53.0: ✅ Pre-registered (list-based)
# 4.54.0: ✅ Pre-registered (layers-based)
# 4.54.1: ✅ Pre-registered (layers-based)
# 4.55.0: ✅ Pre-registered (layers-based)
# 4.55.1: ✅ Pre-registered (layers-based)
# 4.56.0: ❌ Not registered (layers-based)
# 4.56.1: ❌ Not registered (layers-based)
# 4.57.0: ❌ Not registered (layers-based)
# 4.57.1: ❌ Not registered (layers-based)
# 4.57.2: ❌ Not registered (layers-based)
# 4.57.3: ❌ Not registered (layers-based)
# 4.57.4: ❌ Not registered (layers-based)
# 4.57.5: ❌ Not registered (layers-based)
# 4.57.6: ❌ Not registered (layers-based)
# 5.0.0: ❌ Not registered (layers-based)
#
##################################################################################

##################################################################################
#
# All _flatten_* / _unflatten_* helpers are defined at module scope (not inside
# functions) so that torch pytree serialization can locate them by name.
#
# Convention for every cache type:
#   _flatten_<type>           -> (children, aux_data)   [main pytree API]
#   _unflatten_<type>         -> reconstructed object
#   _flatten_with_keys_<type> -> keyed children list     [for pytree.register_pytree_node]
#   _flatten_<type>_for_fx    -> flat list                [for fx_pytree.register_pytree_flatten_spec]\
#
##################################################################################


# ---------------------------------------------------------------------------
# StaticCache
# ---------------------------------------------------------------------------


def _flatten_static_cache(cache) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
    children = (cache.layers,)
    aux_data = {
        "layer_class_to_replicate": getattr(cache, "layer_class_to_replicate", None),
        "offloading": getattr(cache, "offloading", False),
    }
    return children, aux_data


def _unflatten_static_cache(children: Tuple[Any, ...], aux_data: Dict[str, Any]):
    from transformers.cache_utils import StaticCache

    instance = StaticCache.__new__(StaticCache)
    (instance.layers,) = children
    for key, value in aux_data.items():
        setattr(instance, key, value)
    return instance


def _flatten_with_keys_static_cache(cache):
    children, aux_data = _flatten_static_cache(cache)
    return [(pytree.GetAttrKey("layers"), children[0])], aux_data


def _flatten_static_cache_for_fx(cache, spec):
    children, _ = _flatten_static_cache(cache)
    return list(children)


def register_static_cache():
    # StaticCache uses a layers-based structure only when StaticLayer is available
    # (transformers >= ~4.57).  On older versions _flatten_static_cache with non-existent
    # '.layers' attribute must have been pre-registered to pytree, so we skip registration.
    try:
        from transformers.cache_utils import StaticCache, StaticLayer  # noqa: F401
    except ImportError:
        logger = logging.getLogger(__name__)
        logger.debug(
            "StaticCache / StaticLayer not available in this transformers version; "
            "skipping StaticCache pytree registration."
        )
        return

    try:
        pytree.register_pytree_node(
            StaticCache,
            _flatten_static_cache,
            _unflatten_static_cache,
            serialized_type_name=f"{StaticCache.__module__}.{StaticCache.__name__}",
            flatten_with_keys_fn=_flatten_with_keys_static_cache,
        )
        fx_pytree.register_pytree_flatten_spec(
            StaticCache, _flatten_static_cache_for_fx
        )
    except ValueError as e:
        logger = logging.getLogger(__name__)
        logger.debug(f"StaticCache is already registered as pytree flattenable. {e}")


# ---------------------------------------------------------------------------
# StaticLayer
# ---------------------------------------------------------------------------


def _flatten_static_layer(layer) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
    """Split a StaticLayer into (tensor children, static metadata)."""
    if not getattr(layer, "is_initialized", False):
        raise ValueError(
            f"{layer} cannot be flattened. StaticLayer must be initialized "
            "with tensors of a specific shape before use with torch.export."
        )

    # 1. convert tracing tensors
    children = (layer.keys, layer.values)

    # 2. extract aux with compatibility
    aux_data: Dict[str, Any] = {
        "max_cache_len": getattr(layer, "max_cache_len", 0),
        "is_initialized": getattr(layer, "is_initialized", False),
        "dtype": getattr(layer, "dtype", None),
        "device": getattr(layer, "device", None),
        "max_batch_size": getattr(
            layer, "max_batch_size", getattr(layer, "batch_size", 0)
        ),
        "num_heads": getattr(layer, "num_heads", 0),
        "head_dim": getattr(layer, "head_dim", 0),
    }
    return children, aux_data


def _unflatten_static_layer(children: Tuple[Any, ...], aux_data: Dict[str, Any]):
    """Reconstruct a StaticLayer from flattened data."""
    from transformers.cache_utils import StaticLayer

    keys, values = children

    # avoid __init__ signature mismatch error, depending on torch version
    obj = object.__new__(StaticLayer)

    for key, value in aux_data.items():
        if value is not None:
            setattr(obj, key, value)

    # For compatibility
    if not hasattr(obj, "batch_size") and "max_batch_size" in aux_data:
        setattr(obj, "batch_size", aux_data["max_batch_size"])

    obj.keys = keys
    obj.values = values

    return obj


def _flatten_with_keys_static_layer(layer):
    children, aux_data = _flatten_static_layer(layer)

    # In casse of GetAttrKey does not exist in old torch version
    KeyClass = getattr(pytree, "GetAttrKey", getattr(pytree, "MappingKey", None))

    if KeyClass is None:
        # In case of very old torch version
        return [("keys", children[0]), ("values", children[1])], aux_data

    return [
        (KeyClass("keys"), children[0]),
        (KeyClass("values"), children[1]),
    ], aux_data


def _flatten_static_layer_for_fx(layer, spec):
    children, _ = _flatten_static_layer(layer)
    return list(children)


def register_static_layer():
    try:
        from transformers.cache_utils import StaticLayer
    except ImportError:
        logger = logging.getLogger(__name__)
        logger.debug(
            "StaticLayer not available in this transformers version; "
            "skipping StaticLayer pytree registration."
        )
        return

    try:
        pytree.register_pytree_node(
            StaticLayer,
            _flatten_static_layer,
            _unflatten_static_layer,
            serialized_type_name=f"{StaticLayer.__module__}.{StaticLayer.__name__}",
            flatten_with_keys_fn=_flatten_with_keys_static_layer,
        )
        fx_pytree.register_pytree_flatten_spec(
            StaticLayer, _flatten_static_layer_for_fx
        )
    except ValueError as e:
        logger = logging.getLogger(__name__)
        logger.debug(f"StaticLayer is already registered as pytree flattenable. {e}")


# ---------------------------------------------------------------------------
# DynamicLayer
# ---------------------------------------------------------------------------


def _flatten_dynamic_layer(layer) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
    if not layer.is_initialized:
        raise ValueError(
            f"{layer} cannot be flattened. DynamicLayer must be initialized "
            "with tensors of a specific shape before use with torch.export."
        )
    children = (layer.keys, layer.values)
    aux_data: Dict[str, Any] = {
        "is_initialized": layer.is_initialized,
        "dtype": layer.keys.dtype,
        "device": layer.keys.device,
    }
    return children, aux_data


def _unflatten_dynamic_layer(children: Tuple[Any, ...], aux_data: Dict[str, Any]):
    from transformers.cache_utils import DynamicLayer

    keys, values = children
    obj = DynamicLayer()
    obj.keys = keys
    obj.values = values
    obj.is_initialized = aux_data["is_initialized"]
    obj.dtype = aux_data["dtype"]
    obj.device = aux_data["device"]
    return obj


def _flatten_with_keys_dynamic_layer(layer):
    children, aux_data = _flatten_dynamic_layer(layer)
    return [
        (pytree.GetAttrKey("keys"), children[0]),
        (pytree.GetAttrKey("values"), children[1]),
    ], aux_data


def _flatten_dynamic_layer_for_fx(layer, spec):
    children, _ = _flatten_dynamic_layer(layer)
    return list(children)


def register_dynamic_layer():
    try:
        from transformers.cache_utils import DynamicLayer
    except ImportError:
        logger = logging.getLogger(__name__)
        logger.debug(
            "DynamicLayer not available in this transformers version; "
            "skipping DynamicLayer pytree registration."
        )
        return

    try:
        pytree.register_pytree_node(
            DynamicLayer,
            _flatten_dynamic_layer,
            _unflatten_dynamic_layer,
            serialized_type_name=f"{DynamicLayer.__module__}.{DynamicLayer.__name__}",
            flatten_with_keys_fn=_flatten_with_keys_dynamic_layer,
        )
        fx_pytree.register_pytree_flatten_spec(
            DynamicLayer, _flatten_dynamic_layer_for_fx
        )
    except ValueError as e:
        logger = logging.getLogger(__name__)
        logger.debug(f"DynamicLayer is already registered as pytree flattenable. {e}")


# ---------------------------------------------------------------------------
# DynamicCache
# ---------------------------------------------------------------------------


def _flatten_dynamic_cache(cache) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
    children = (cache.layers,)
    aux_data = {
        "layer_class_to_replicate": getattr(cache, "layer_class_to_replicate", None),
        "offloading": getattr(cache, "offloading", False),
    }
    return children, aux_data


def _unflatten_dynamic_cache(children: Tuple[Any, ...], aux_data: Dict[str, Any]):
    from transformers.cache_utils import DynamicCache

    instance = DynamicCache.__new__(DynamicCache)
    (instance.layers,) = children
    for key, value in aux_data.items():
        setattr(instance, key, value)
    return instance


def _flatten_with_keys_dynamic_cache(cache):
    children, aux_data = _flatten_dynamic_cache(cache)
    return [(pytree.GetAttrKey("layers"), children[0])], aux_data


def _flatten_dynamic_cache_for_fx(cache, spec):
    children, _ = _flatten_dynamic_cache(cache)
    return list(children)


# Legacy flatten/unflatten for transformers versions that do not have
# DynamicLayer (e.g. <= 4.52.x), which store tensors directly in
# key_cache / value_cache instead of the Layer-based cache.layers structure.
def _flatten_dynamic_cache_legacy(cache) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
    children = (cache.key_cache, cache.value_cache)
    aux_data: Dict[str, Any] = {}
    return children, aux_data


def _unflatten_dynamic_cache_legacy(
    children: Tuple[Any, ...], aux_data: Dict[str, Any]
):
    from transformers.cache_utils import DynamicCache

    key_cache, value_cache = children
    cache = DynamicCache()
    cache.key_cache = key_cache
    cache.value_cache = value_cache
    return cache


def _flatten_with_keys_dynamic_cache_legacy(cache):
    children, aux_data = _flatten_dynamic_cache_legacy(cache)
    return [
        (pytree.GetAttrKey("key_cache"), children[0]),
        (pytree.GetAttrKey("value_cache"), children[1]),
    ], aux_data


def _flatten_dynamic_cache_for_fx_legacy(cache, spec):
    children, _ = _flatten_dynamic_cache_legacy(cache)
    return list(children)


def register_dynamic_cache():
    """Register DynamicCache as a pytree node.

    Two layouts exist across transformers versions:

    * **Layer-based** (newer, requires ``DynamicLayer`` to be importable):
      ``cache.layers`` is a list of ``DynamicLayer`` objects; each layer
      holds ``keys`` and ``values`` tensors.  Both ``DynamicCache`` and
      ``DynamicLayer`` must be registered as pytree nodes for
      ``torch.export`` to trace through the cache.

    * **Legacy** (older, e.g. transformers <= 4.52.x):
      The cache stores tensors directly in ``cache.key_cache`` and
      ``cache.value_cache`` lists.  No ``DynamicLayer`` class exists.

    The correct layout is detected by checking whether ``DynamicLayer`` can
    be imported rather than by comparing version strings, which have proven
    unreliable across patch releases.
    """
    if not is_transformers_installed:  # type: ignore[truthy-function]
        raise ImportError("transformers package is not installed")

    from transformers.cache_utils import DynamicCache

    # Feature-detect the Layer-based layout.
    try:
        from transformers.cache_utils import DynamicLayer as _DL  # noqa: F401

        _has_dynamic_layer = True
    except ImportError:
        _has_dynamic_layer = False

    if not _has_dynamic_layer:
        # Legacy layout: flatten key_cache / value_cache directly.
        try:
            pytree.register_pytree_node(
                DynamicCache,
                _flatten_dynamic_cache_legacy,
                _unflatten_dynamic_cache_legacy,
                serialized_type_name=f"{DynamicCache.__module__}.{DynamicCache.__name__}",
                flatten_with_keys_fn=_flatten_with_keys_dynamic_cache_legacy,
            )
            fx_pytree.register_pytree_flatten_spec(
                DynamicCache, _flatten_dynamic_cache_for_fx_legacy
            )
        except ValueError as e:
            logger = logging.getLogger(__name__)
            logger.debug(
                f"DynamicCache is already registered as pytree flattenable. {e}"
            )
        return

    try:
        pytree.register_pytree_node(
            DynamicCache,
            _flatten_dynamic_cache,
            _unflatten_dynamic_cache,
            serialized_type_name=f"{DynamicCache.__module__}.{DynamicCache.__name__}",
            flatten_with_keys_fn=_flatten_with_keys_dynamic_cache,
        )
        fx_pytree.register_pytree_flatten_spec(
            DynamicCache, _flatten_dynamic_cache_for_fx
        )
    except ValueError as e:
        logger = logging.getLogger(__name__)
        logger.debug(f"DynamicCache is already registered as pytree flattenable. {e}")


# ---------------------------------------------------------------------------
# EncoderDecoderCache
# ---------------------------------------------------------------------------


def _flatten_encoder_decoder_cache(cache) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
    children = (cache.self_attention_cache, cache.cross_attention_cache)
    aux_data: Dict[str, Any] = {}
    return children, aux_data


def _unflatten_encoder_decoder_cache(
    children: Tuple[Any, ...], aux_data: Dict[str, Any]
):
    from transformers.cache_utils import EncoderDecoderCache

    self_cache, cross_cache = children
    return EncoderDecoderCache(self_cache, cross_cache)


def _flatten_with_keys_encoder_decoder_cache(cache):
    children, aux_data = _flatten_encoder_decoder_cache(cache)
    return [
        (pytree.GetAttrKey("self_attention_cache"), children[0]),
        (pytree.GetAttrKey("cross_attention_cache"), children[1]),
    ], aux_data


def _flatten_encoder_decoder_cache_for_fx(cache, spec):
    children, _ = _flatten_encoder_decoder_cache(cache)
    return list(children)


def register_encoder_decoder_cache():
    from transformers.cache_utils import EncoderDecoderCache

    try:
        pytree.register_pytree_node(
            EncoderDecoderCache,
            _flatten_encoder_decoder_cache,
            _unflatten_encoder_decoder_cache,
            serialized_type_name=(
                f"{EncoderDecoderCache.__module__}.{EncoderDecoderCache.__name__}"
            ),
            flatten_with_keys_fn=_flatten_with_keys_encoder_decoder_cache,
        )
        fx_pytree.register_pytree_flatten_spec(
            EncoderDecoderCache, _flatten_encoder_decoder_cache_for_fx
        )
    except ValueError as e:
        logger = logging.getLogger(__name__)
        logger.debug(
            f"EncoderDecoderCache is already registered as pytree flattenable. {e}"
        )



# Copyright (c) 2025 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, TYPE_CHECKING

import torch

from transformers.cache_utils import DynamicCache
from transformers.models.llama.modeling_llama import LlamaAttention


def llama_attention_forward_adapter(
    self: LlamaAttention,
    hidden_states: torch.Tensor,
    position_embeddings: List[torch.Tensor],
    attention_mask: torch.Tensor,
    past_key_value: DynamicCache,
    cache_position: torch.Tensor,
    **kwargs,
):
    # past_key_value is a dict with key_cache and value_cache.
    # It needs to be decomposed for tico and circle which does not know dict.
    key_cache = past_key_value.key_cache  # type: ignore[union-attr]
    value_cache = past_key_value.value_cache  # type: ignore[union-attr]
    return (
        torch.ops.circle_custom.attention(
            hidden_states,
            self.q_proj.weight,
            self.k_proj.weight,
            self.v_proj.weight,
            self.o_proj.weight,
            position_embeddings[0],  # cos
            position_embeddings[1],  # sin
            attention_mask,
            key_cache[self.layer_idx],
            value_cache[self.layer_idx],  # Same to value_cache
            cache_position,
        ),
        None,
    )



# Copyright (c) 2025 Samsung Electronics Co., Ltd. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

import numpy as np
import torch
from cffi import FFI


class Interpreter:
    """
    Python wrapper for C++ luci-interperter class in ONE using CFFI.

    This class provides a Python interface to the underlying C++ luci-interpreter class in ONE,
     preserving the original C++ API. Each method corresponds to a method in the C++ class,
    with additional error handling implemented to ensure that C++ exceptions are captured and
    translated into Python errors.

    Note that each method includes `check_for_errors` at the end of the body to catch any C++
     exceptions and translate them into Python exceptions. This ensures that errors in the C++
    library do not cause undefined behavior in Python.
    """

    def __init__(self, circle_binary: bytes):
        self.ffi = FFI()
        self.ffi.cdef(
            """
          typedef struct InterpreterWrapper InterpreterWrapper;

          const char *get_last_error(void);
          void clear_last_error(void);
          InterpreterWrapper *Interpreter_new(const uint8_t *data, const size_t data_size);
          void Interpreter_delete(InterpreterWrapper *intp);
          void Interpreter_interpret(InterpreterWrapper *intp);
          void Interpreter_writeInputTensor(InterpreterWrapper *intp, const int input_idx, const void *data, size_t input_size);
          void Interpreter_readOutputTensor(InterpreterWrapper *intp, const int output_idx, void *output, size_t output_size);
        """
        )
        # TODO Check if one-compiler version is compatible. Whether it has .so file or not for CFFI.
        intp_lib_path = Path("/usr/share/one/lib/libcircle_interpreter_cffi.so")
        if not intp_lib_path.is_file():
            raise RuntimeError("Please install one-compiler for circle inference.")
        self.C = self.ffi.dlopen(str(intp_lib_path))

        # Initialize interpreter
        self.intp = self.C.Interpreter_new(circle_binary, len(circle_binary))
        self.check_for_errors()

    def delete(self):
        self.C.Interpreter_delete(self.intp)
        self.check_for_errors()

    def interpret(self):
        self.C.Interpreter_interpret(self.intp)
        self.check_for_errors()

    def writeInputTensor(self, input_idx: int, input_data: torch.Tensor):
        input_as_numpy = input_data.numpy()
        # cffi.from_buffer() only accepts C-contiguous array.
        input_as_numpy = np.ascontiguousarray(input_as_numpy)
        c_input = self.ffi.from_buffer(input_as_numpy)
        self.C.Interpreter_writeInputTensor(
            self.intp, input_idx, c_input, input_data.nbytes
        )
        self.check_for_errors()

    def readOutputTensor(self, output_idx: int, output: np.ndarray):
        c_output = self.ffi.from_buffer(output)
        self.C.Interpreter_readOutputTensor(
            self.intp, output_idx, c_output, output.nbytes
        )
        self.check_for_errors()

    def check_for_errors(self):
        error_message = self.ffi.string(self.C.get_last_error()).decode("utf-8")
        if error_message:
            self.C.clear_last_error()
            raise RuntimeError(f"C++ Exception: {error_message}")

    def __del__(self):
        self.delete()
